"""
Test the check_parameters module

:author: Michel Bierlaire
:date: Thu Dec  1 16:35:18 2022

"""
import unittest
import biogeme.check_parameters as cp


class TestCheckParameters(unittest.TestCase):
    """Tests for the biogeme.check_parameters module"""

    def test_zero_one(self):
        diag, _ = cp.zero_one(0.5)
        self.assertEqual(diag, True)
        diag, _ = cp.zero_one(-0.5)
        self.assertEqual(diag, False)
        diag, _ = cp.zero_one(1.5)
        self.assertEqual(diag, False)

    def test_is_number(self):
        diag, _ = cp.is_number(1.2)
        self.assertEqual(diag, True)
        diag, _ = cp.is_number(1)
        self.assertEqual(diag, True)
        diag, _ = cp.is_number(-1)
        self.assertEqual(diag, True)
        diag, _ = cp.is_number(-1.2e-4)
        self.assertEqual(diag, True)
        diag, _ = cp.is_number('a_string')
        self.assertEqual(diag, False)

    def test_is_positive(self):
        diag, _ = cp.is_positive(1.2)
        self.assertEqual(diag, True)
        diag, _ = cp.is_positive(0)
        self.assertEqual(diag, False)
        diag, _ = cp.is_positive(-1.2)
        self.assertEqual(diag, False)

    def test_is_integer(self):
        diag, _ = cp.is_integer(1)
        self.assertEqual(diag, True)
        diag, _ = cp.is_integer(1.2)
        self.assertEqual(diag, False)

    def test_check_algo_name(self):
        diag, _ = cp.check_algo_name('scipy')
        self.assertEqual(diag, True)
        diag, _ = cp.check_algo_name('any_string')
        self.assertEqual(diag, False)

    def test_boolean(self):
        diag, _ = cp.is_boolean(True)
        self.assertEqual(diag, True)
        diag, _ = cp.is_boolean(1)
        self.assertEqual(diag, False)
        diag, _ = cp.is_boolean('True')
        self.assertEqual(diag, False)

    def test_is_non_negative(self):
        diag, _ = cp.is_non_negative(1.2)
        self.assertEqual(diag, True)
        diag, _ = cp.is_non_negative(0)
        self.assertEqual(diag, True)
        diag, _ = cp.is_non_negative(-1.2)
        self.assertEqual(diag, False)



"""
Test the results module

:author: Michel Bierlaire
:date: Wed Aug 25 09:41:24 2021
"""
# Bug in pylint
# pylint: disable=no-member
#
# Too constraining
# pylint: disable=invalid-name, too-many-instance-attributes
#
# Not needed in test
# pylint: disable=missing-function-docstring, missing-class-docstring

import unittest
import pandas as pd
import biogeme.biogeme as bio
import biogeme.database as db
import biogeme.results as res
from biogeme.expressions import Beta, Variable, exp


class TestResults(unittest.TestCase):
    def setUp(self):
        df = pd.DataFrame(
            {
                'Person': [1, 1, 1, 2, 2],
                'Exclude': [0, 0, 1, 0, 1],
                'Variable1': [1, 2, 3, 4, 5],
                'Variable2': [10, 20, 30, 40, 50],
                'Choice': [1, 2, 3, 1, 2],
                'Av1': [0, 1, 1, 1, 1],
                'Av2': [1, 1, 1, 1, 1],
                'Av3': [0, 1, 1, 1, 1],
            }
        )
        myData = db.Database('test', df)

        Variable1 = Variable('Variable1')
        Variable2 = Variable('Variable2')
        beta1 = Beta('beta1', -1.0, -3, 3, 0)
        beta2 = Beta('beta2', 2.0, -3, 10, 0)
        likelihood = (
            -(beta1**2) * Variable1 - exp(beta2 * beta1) * Variable2 - beta2**4
        )
        simul = beta1 / Variable1 + beta2 / Variable2
        dictOfExpressions = {
            'loglike': likelihood,
            'beta1': beta1,
            'simul': simul,
        }
        my_biogeme = bio.BIOGEME(myData, dictOfExpressions)
        my_biogeme.generate_html = False
        my_biogeme.generate_pickle = False
        my_biogeme.saveIterations = False
        my_biogeme.modelName = 'simpleExample'
        my_biogeme.bootstrap_samples = 10
        self.results = my_biogeme.estimate(run_bootstrap=True)

    def test_calcPValue(self):
        p = res.calcPValue(1.96)
        self.assertAlmostEqual(p, 0.05, 2)

    def test_isBoundActive(self):
        beta1 = res.beta('beta1', -1, (-1, 1))
        self.assertTrue(beta1.is_bound_active())
        beta2 = res.beta('beta2', 0, (-1, 1))
        self.assertFalse(beta2.is_bound_active())


if __name__ == '__main__':
    unittest.main()



"""File b03simulation.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 21:04:33 2023

We use an estimated model to perform various simulations
"""
import sys
import time
import pandas as pd
from biogeme import models
import biogeme.biogeme as bio
import biogeme.exceptions as excep
import biogeme.results as res
from optima_data import database, normalized_weight
from scenarios import scenario

# Obtain the specification for the default scenario
V, nests, _, _ = scenario()

V_PT = V[0]
V_CAR = V[1]
V_SM = V[2]

# Obtain the expression for the choice probability of each alternative
prob_PT = models.nested(V, None, nests, 0)
prob_CAR = models.nested(V, None, nests, 1)
prob_SM = models.nested(V, None, nests, 2)

# Read the estimation results from the file
try:
    results = res.bioResults(pickleFile='b02estimation.pickle')
except excep.BiogemeError:
    sys.exit(
        'Run first the script b02simulation.py '
        'in order to generate the '
        'file b02estimation.pickle.'
    )


# We now simulate various expressions on the database, and store the
# results in a Pandas dataframe

start_time = time.time()
simulate_formulas = {
    'weight': normalized_weight.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
    'Utility PT': V_PT.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
    'Utility car': V_CAR.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
    'Utility SM': V_SM.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
    'Prob. PT': prob_PT.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
    'Prob. car': prob_CAR.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
    'Prob. SM': prob_SM.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
}

simulated_values = pd.DataFrame.from_dict(
    simulate_formulas,
)
print(
    f'--- Execution time with getValue_c: '
    f'{time.time() - start_time:.2f} seconds ---'
)

# We now perform the same simulation using Biogeme. The results are be
# identical, but the execution time is faster. Indeed, Biogeme
# recycles calculations performed for one expression for the other
# expressions.

# A dictionary with the requested expression must be provided to Biogeme

simulate = {
    'weight': normalized_weight,
    'Utility PT': V_PT,
    'Utility car': V_CAR,
    'Utility SM': V_SM,
    'Prob. PT': prob_PT,
    'Prob. car': prob_CAR,
    'Prob. SM': prob_SM,
}

start_time = time.time()
the_biogeme = bio.BIOGEME(database, simulate)
biogeme_simulation = the_biogeme.simulate(results.getBetaValues())
print(
    f'--- Execution time with Biogeme:    '
    f'{time.time() - start_time:.2f} seconds ---'
)

# Let's print the two results, to show that they are identical

print('Results without Biogeme')
print(simulated_values)
print('Results with Biogeme')
print(biogeme_simulation)



""" Interface for the use of the loggers in Biogeme

:author: Michel Bierlaire
:date: Thu Mar 23 17:53:22 2023

"""
import logging

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

loggers = (
    logging.getLogger('biogeme'),
    logging.getLogger('biogeme_optimization'),
)


def get_screen_logger(level=WARNING):
    """Obtain a screen logger

    :param level: level of verbosity of the logger
    :type level: int
    """
    for logger in loggers:
        # This has to be set to the lower level: DEBUG so that it does not
        # supersed the handler
        logger.setLevel(DEBUG)
        formatter_debug = logging.Formatter(
            '[%(levelname)s] %(asctime)s %(message)s <%(filename)s:%(lineno)d>'
        )
        formatter_normal = logging.Formatter('%(message)s ')
        formatter = formatter_debug if level == DEBUG else formatter_normal
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(level)
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)
    return loggers[0]


def get_file_logger(filename, level=WARNING):
    """Obtain a file logger

    :param filename: name of the file. Extension .log recommended.
    :type filename: str

    :param level: level of verbosity of the logger
    :type level: int
    """
    for logger in loggers:
        # This has to be set to the lower level: DEBUG so that it does not
        # supersed the handler
        logger.setLevel(DEBUG)
        formatter = logging.Formatter(
            '[%(levelname)s] %(asctime)s %(message)s <%(filename)s:%(lineno)d>'
        )
        file_handler = logging.FileHandler(filename)
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    return loggers[0]



"""Implements some useful functions

:author: Michel Bierlaire

:date: Sun Apr 14 10:46:10 2019

"""

import sys
import logging
from itertools import product
from os import path
import shutil
from collections import defaultdict
from typing import NamedTuple
import tempfile
import uuid
import numpy as np
import pandas as pd
from scipy.stats import chi2
import biogeme.exceptions as excep

logger = logging.getLogger(__name__)


class LRTuple(NamedTuple):
    message: str
    statistic: float
    threshold: float


def findiff_g(the_function, x):
    """Calculates the gradient of a function :math:`f` using finite differences

    :param the_function: A function object that takes a vector as an
                        argument, and returns a tuple. The first
                        element of the tuple is the value of the
                        function :math:`f`. The other elements are not
                        used.
    :type the_function: function

    :param x: argument of the function
    :type x: numpy.array

    :return: numpy vector, same dimension as x, containing the gradient
       calculated by finite differences.
    :rtype: numpy.array

    """
    x = x.astype(float)
    tau = 0.0000001
    n = len(x)
    g = np.zeros(n)
    f = the_function(x)[0]
    for i in range(n):
        xi = x.item(i)
        xp = x.copy()
        if abs(xi) >= 1:
            s = tau * xi
        elif xi >= 0:
            s = tau
        else:
            s = -tau
        xp[i] = xi + s
        fp = the_function(xp)[0]
        g[i] = (fp - f) / s
    return g


def findiff_H(the_function, x):
    """Calculates the hessian of a function :math:`f` using finite differences

    :param the_function: A function object that takes a vector as an
                        argument, and returns a tuple. The first
                        element of the tuple is the value of the
                        function :math:`f`, and the second is the
                        gradient of the function.  The other elements
                        are not used.

    :type the_function: function

    :param x: argument of the function
    :type x: numpy.array

    :return: numpy matrix containing the hessian calculated by
             finite differences.
    :rtype: numpy.array

    """
    tau = 1.0e-7
    n = len(x)
    H = np.zeros((n, n))
    g = the_function(x)[1]
    eye = np.eye(n, n)
    for i in range(n):
        xi = x.item(i)
        if abs(xi) >= 1:
            s = tau * xi
        elif xi >= 0:
            s = tau
        else:
            s = -tau
        ei = eye[i]
        gp = the_function(x + s * ei)[1]
        H[:, i] = (gp - g).flatten() / s
    return H


def checkDerivatives(the_function, x, names=None, logg=False):
    """Verifies the analytical derivatives of a function by comparing
    them with finite difference approximations.

    :param the_function: A function object that takes a vector as an argument,
        and returns a tuple:

        - The first element of the tuple is the value of the
          function :math:`f`,
        - the second is the gradient of the function,
        - the third is the hessian.

    :type the_function: function

    :param x: arguments of the function
    :type x: numpy.array

    :param names: the names of the entries of x (for reporting).
    :type names: list(string)
    :param logg: if True, messages will be displayed.
    :type logg: bool


    :return: tuple f, g, h, gdiff, hdiff where

          - f is the value of the function at x,
          - g is the analytical gradient,
          - h is the analytical hessian,
          - gdiff is the difference between the analytical gradient
            and the finite difference approximation
          - hdiff is the difference between the analytical hessian
            and the finite difference approximation

    :rtype: float, numpy.array,numpy.array,  numpy.array,numpy.array

    """
    x = np.array(x, dtype=float)
    f, g, h = the_function(x)
    g_num = findiff_g(the_function, x)
    gdiff = g - g_num
    if logg:
        if names is None:
            names = [f'x[{i}]' for i in range(len(x))]
        logger.info('x\t\tGradient\tFinDiff\t\tDifference')
        for k, v in enumerate(gdiff):
            logger.info(f'{names[k]:15}\t{g[k]:+E}\t{g_num[k]:+E}\t{v:+E}')

    h_num = findiff_H(the_function, x)
    hdiff = h - h_num
    if logg:
        logger.info('Row\t\tCol\t\tHessian\tFinDiff\t\tDifference')
        for row in range(len(hdiff)):
            for col in range(len(hdiff)):
                logger.info(
                    f'{names[row]:15}\t{names[col]:15}\t{h[row,col]:+E}\t'
                    f'{h_num[row,col]:+E}\t{hdiff[row,col]:+E}'
                )
    return f, g, h, gdiff, hdiff


def get_prime_numbers(n):
    """Get a given number of prime numbers

    :param n: number of primes that are requested
    :type n: int

    :return: array with prime numbers
    :rtype: list(int)

    :raise BiogemeError: if the requested number is non positive or a float

    """
    total = 0
    upper_bound = 100
    if n <= 0:
        raise excep.BiogemeError(f'Incorrect number: {n}')

    while total < n:
        upper_bound *= 10
        primes = calculate_prime_numbers(upper_bound)
        total = len(primes)
    try:
        return primes[0:n]
    except TypeError as e:
        raise excep.BiogemeError(f'Incorrect number: {n}') from e


def calculate_prime_numbers(upper_bound):
    """Calculate prime numbers

    :param upper_bound: prime numbers up to this value will be computed
    :type upper_bound: int

    :return: array with prime numbers
    :rtype: list(int)

    :raise BiogemeError: if the upper_bound is incorrectly defined
        (negative number, e.g.)

    >>> tools.calculate_prime_numbers(10)
    [2, 3, 5, 7]

    """
    if upper_bound < 0:
        raise excep.BiogemeError(f'Incorrect value: {upper_bound}')
    try:
        mywork = list(range(0, upper_bound + 1))
    except TypeError as e:
        raise excep.BiogemeError(f'Incorrect value: {upper_bound}') from e

    try:
        largest = int(np.ceil(np.sqrt(float(upper_bound))))
    except ValueError as e:
        raise excep.BiogemeError(f'Incorrect value: {upper_bound}') from e

    # Remove all multiples
    for i in range(2, largest + 1):
        if mywork[i] != 0:
            for k in range(2 * i, upper_bound + 1, i):
                mywork[k] = 0
    # Gather non zero entries, which are the prime numbers
    myprimes = []
    for i in range(1, upper_bound + 1):
        if mywork[i] != 0 and mywork[i] != 1:
            myprimes += [mywork[i]]

    return myprimes


def countNumberOfGroups(df, column):
    """
    This function counts the number of groups of same value in a column.
    For instance: 1,2,2,3,3,3,4,1,1  would give 5.

    Example::

        >>>df = pd.DataFrame({'ID': [1, 1, 2, 3, 3, 1, 2, 3],
                              'value':[1000,
                                       2000,
                                       3000,
                                       4000,
                                       5000,
                                       5000,
                                       10000,
                                       20000]})
        >>>tools.countNumberOfGroups(df,'ID')
        6

        >>>tools.countNumberOfGroups(df,'value')
        7

    """
    df['_biogroups'] = (df[column] != df[column].shift(1)).cumsum()
    result = len(df['_biogroups'].unique())
    df.drop(columns=['_biogroups'], inplace=True)
    return result


def likelihood_ratio_test(model1, model2, significance_level=0.05):
    """This function performs a likelihood ratio test between a
    restricted and an unrestricted model.

    :param model1: the final loglikelihood of one model, and the number of
                   estimated parameters.
    :type model1: tuple(float, int)

    :param model2: the final loglikelihood of the other model, and
                   the number of estimated parameters.
    :type model2: tuple(float, int)

    :param significance_level: level of significance of the test. Default: 0.05
    :type significance_level: float

    :return: a tuple containing:

                  - a message with the outcome of the test
                  - the statistic, that is minus two times the difference
                    between the loglikelihood  of the two models
                  - the threshold of the chi square distribution.

    :rtype: LRTuple(str, float, float)

    :raise BiogemeError: if the unrestricted model has a lower log
        likelihood than the restricted model.

    """

    loglike_m1, df_m1 = model1
    loglike_m2, df_m2 = model2
    if loglike_m1 > loglike_m2:
        if df_m1 < df_m2:
            raise excep.BiogemeError(
                f'The unrestricted model {model2} has a '
                f'lower log likelihood than the restricted one {model1}'
            )
        loglike_ur = loglike_m1
        loglike_r = loglike_m2
        df_ur = df_m1
        df_r = df_m2
    else:
        if df_m1 >= df_m2:
            raise excep.BiogemeError(
                f'The unrestricted model {model1} has a '
                f'lower log likelihood than the restricted one {model2}'
            )
        loglike_ur = loglike_m2
        loglike_r = loglike_m1
        df_ur = df_m2
        df_r = df_m1

    stat = -2 * (loglike_r - loglike_ur)
    chi_df = df_ur - df_r
    threshold = chi2.ppf(1 - significance_level, chi_df)
    if stat <= threshold:
        final_msg = f'H0 cannot be rejected at level {100*significance_level:.1f}%'
    else:
        final_msg = f'H0 can be rejected at level {100*significance_level:.1f}%'
    return LRTuple(message=final_msg, statistic=stat, threshold=threshold)


def flatten_database(df, merge_id, row_name=None, identical_columns=None):
    """Combine several rows of a Pandas database into one.
    For instance, consider the following database::

           ID  Age  Cost   Name
        0   1   23    34  Item3
        1   1   23    45  Item4
        2   1   23    12  Item7
        3   2   45    65  Item3
        4   2   45    34  Item7



    If row_name is 'Name', the function generates the same data in the
    following format::

            Age  Item3_Cost  Item4_Cost  Item7_Cost
        ID
        1    23          34        45.0          12
        2    45          65         NaN          34


    If row_name is None, the function generates the same data in the
    following format::

            Age  1_Cost 1_Name  2_Cost 2_Name  3_Cost 3_Name
        ID
        1    23      34  Item3      45  Item4    12.0  Item7
        2    45      65  Item3      34  Item7     NaN    NaN

    :param df: initial data frame
    :type df: pandas.DataFrame

    :param merge_id: name of the column that identifies rows that
        should be merged. In the above example: 'ID'
    :type merge_id: str

    :param row_name: name of the columns that provides the name of the
        rows in the new dataframe. In the example above: 'Name'. If
        None, the rows are numbered sequentially.
    :type row_name: str

    :param identical_columns: name of the columns that contain
        identical values across the rows of a group. In the example
        above: ['Age']. If None, these columns are automatically
        detected. On large database, there may be a performance issue.
    :type identical_columns: list(str)

    :return: reformatted database
    :rtype: pandas.DataFrame

    """
    grouped = df.groupby(by=merge_id)
    all_columns = set(df.columns)

    def are_values_identical(col):
        """This function checks if all the values in a column
        are identical

        :param col: the column
        :type col: pandas.Series

        :return: True if all values are identical. False otherwise.
        :rtype: bool
        """

        return (col.iloc[0] == col).all(0)

    def get_varying_cols(g):
        """This functions returns the name of all columns
        that have constant values within each group of data.

        :param g: group of data
        :type g: pandas.DataFrame

        :return: name of all columns that have constant values
            within each group of data.
        :rtype: set(str)
        """
        return {colname for colname, col in g.items() if not are_values_identical(col)}

    if identical_columns is None:
        all_varying_cols = grouped.apply(get_varying_cols)
        varying_columns = set.union(*all_varying_cols)
        identical_columns = list(all_columns - varying_columns)
        varying_columns = list(varying_columns)
    else:
        identical_columns = set(identical_columns)
        identical_columns.add(merge_id)
        varying_columns = list(all_columns - identical_columns)

    # Take the first row for columns that are identical
    if identical_columns:
        common_data = df[list(identical_columns)].drop_duplicates(
            merge_id, keep='first'
        )
        common_data.index = common_data[merge_id]
    # Treat the other columns
    grouped_varying = df[[merge_id] + list(varying_columns)].groupby(by=merge_id)

    def treat(x):
        """Treat a group of data.

        :param x: group of data
        :type x: pandas.DataFrame

        :return: the same data organized in one row, with proper column names
        :rtype: pandas.DataFrame

        :raise BiogemeError:  if there are duplicates in the name of
        the row. Indeed, in that case, they cannot be used to name the
        new columns.
        """
        if not are_values_identical(x[merge_id]):
            err_msg = f'Group has different IDs: {x[merge_id]}. ' f'Rows id: {x.index}'
            raise excep.BiogemeError(err_msg)
        if row_name is not None and not x[row_name].is_unique:
            err_msg = (
                f'Entries in column [{row_name}] are not unique. '
                f'This column cannot be used to name the new '
                f'columns:\n{x[[row_name, merge_id]]}. '
            )
            raise excep.BiogemeError(err_msg)

        the_columns = set(x.columns) - {merge_id}
        if row_name is not None:
            the_columns -= {row_name}
        sorted_list = sorted(list(the_columns))
        first = True
        i = 0
        for _, row in x.iterrows():
            i += 1
            if first:
                all_values = [row[merge_id]]
                all_columns = [merge_id]
                first = False
            name = f'{i}' if row_name is None else row[row_name]
            columns = [f'{name}_{c}' for c in sorted_list]
            all_values.extend([row[c] for c in sorted_list])
            all_columns.extend(columns)
        df = pd.DataFrame([all_values], columns=all_columns)
        return df

    flat_data = grouped_varying.apply(treat)
    flat_data.index = flat_data[merge_id]

    # We remove the column 'merge_id' as it is stored as index.
    if identical_columns:
        return pd.concat([common_data, flat_data], axis='columns').drop(
            columns=[merge_id]
        )
    return flat_data.drop(columns=[merge_id])


class TemporaryFile:
    """Class generating a temporary file, so that the user does not
    bother about its location, or even its name

    Example::

        with TemporaryFile() as filename:
            with open(filename, 'w') as f:
                print('stuff', file=f)
    """

    def __enter__(self, name=None):
        self.dir = tempfile.mkdtemp()
        name = str(uuid.uuid4()) if name is None else name
        return path.join(self.dir, name)

    def __exit__(self, ctype, value, traceback):
        """Destroys the temporary directory"""
        shutil.rmtree(self.dir)


class ModelNames:
    """Class generating model names from unique configuration string"""

    def __init__(self, prefix='Model'):
        self.prefix = prefix
        self.dict_of_names = {}
        self.current_number = 0

    def __call__(self, the_id):
        """Get a short model name from a unique ID

        :param the_id: Id of the model
        :type the_id: str (or anything that can be used as a key for a dict)
        """
        the_name = self.dict_of_names.get(the_id)
        if the_name is None:
            the_name = f'{self.prefix}_{self.current_number:06d}'
            self.current_number += 1
            self.dict_of_names[the_id] = the_name
        return the_name


def generate_unique_ids(list_of_ids):
    """If there are duplicates in the list, a new list is generated
    where there are renamed to obtain a list with unique IDs.

    :param list_of_ids: list of ids
    :type list_of_ids: list[str]

    :return: a dict that maps the unique names with the original name
    """
    counts = defaultdict(int)
    for the_id in list_of_ids:
        counts[the_id] += 1

    results = {}
    for name, count in counts.items():
        if count == 1:
            results[name] = name
        else:
            substitutes = [f'{name}_{i}' for i in range(count)]
            for new_name in substitutes:
                results[new_name] = name
    return results


def unique_product(*iterables, max_memory_mb=1024):
    """Generate the Cartesian product of multiple iterables, keeping
    only the unique entries.  Raises a MemoryError if memory usage
    exceeds the specified threshold.

    :param iterables: Variable number of iterables to compute the
        Cartesian product from.
    :type iterables: Iterable

    :param max_memory_mb: Maximum memory usage in megabytes (default: 1024MB).
    :type max_memory_mb: int

    :return: Yields unique entries from the Cartesian product.
    :rtype: Iterator[tuple]

    """

    MB_TO_BYTES = 1024 * 1024
    max_memory_bytes = max_memory_mb * MB_TO_BYTES  # Convert MB to bytes
    seen = set()  # Set to store seen entries
    total_memory = 0  # Track memory usage

    for items in product(*iterables):
        if items not in seen:
            seen.add(items)
            item_size = sum(sys.getsizeof(item) for item in items)
            total_memory += item_size
            if total_memory > max_memory_bytes:
                raise MemoryError(
                    f'Memory usage exceeded the specified threshold: '
                    f'{total_memory/MB_TO_BYTES:.1f} MB > '
                    f'{max_memory_bytes/MB_TO_BYTES} MB.'
                )
            yield items



"""File specification.py

:author: Michel Bierlaire
:date: Mon Apr 10 12:33:18 2023

Implements a model specification in a multiple expression context (using Catalogs)
"""
import logging
from biogeme_optimization.pareto import SetElement
import biogeme.biogeme as bio
from biogeme import tools
from biogeme.configuration import Configuration
import biogeme.exceptions as excep

logger = logging.getLogger(__name__)


class Specification:
    """Implements a specification"""

    database = None  #: :class:`biogeme.database.Database` object
    all_results = {}  #: dict(str: `biogeme.results.bioResults`)
    expression = None  #: :class:`biogeme.expressions.Expression` object
    """
        function that generates all the objectives:
        fct(bioResults) -> list[floatNone]
    """
    validity = None
    """
    function that checks the validity of the results
    """
    generic_name = 'default_name'  #: short name for file names

    def __init__(self, configuration):
        """Creates a specification from a  configuration

        :param configuration: configuration of the multiple expression
        :type configuration: biogeme.configuration.Configuration
        """
        if not isinstance(configuration, Configuration):
            error_msg = 'Ctor needs an object of type Configuration'
            raise excep.BiogemeError(error_msg)
        self.configuration = configuration
        self.model_names = None
        self._estimate()

    @classmethod
    def from_string_id(cls, configuration_id):
        """Constructor using a configuration"""
        return cls(Configuration.from_string(configuration_id))

    def configure_expression(self):
        """Configure the expression to the current configuration"""
        self.expression.configure_catalogs(self.configuration)

    @classmethod
    def default_specification(cls):
        """Alternative constructor for generate the default specification"""
        cls.expression.reset_expression_selection()
        the_config = cls.expression.current_configuration()
        return cls(the_config)

    @property
    def config_id(self):
        """Defined config_id as a property"""
        return self.configuration.get_string_id()

    @config_id.setter
    def config_id(self, value):
        self.configuration = Configuration.from_string(value)

    def get_results(self):
        """Obtain the estimation results of the specification"""
        the_results = self.all_results.get(self.config_id)
        if the_results is None:
            error_msg = f'No result is available for specification {self.config_id}'
            raise excep.BiogemeError(error_msg)
        return the_results

    def __repr__(self):
        return str(self.config_id)

    def _estimate(self):
        """Estimate the parameter of the current specification, if not already done

        :param quick_estimate: if True, a "quick estimate" is
            performed, in the sense that the final statistics are not
            calculated
        :type quick_estimate: bool
        """
        if self.expression is None:
            error_msg = 'No expression has been provided for the model.'
            raise excep.BiogemeError(error_msg)
        if self.database is None:
            error_msg = 'No database has been provided for the estimation.'
            raise excep.BiogemeError(error_msg)
        if self.model_names is None:
            self.model_names = tools.ModelNames(prefix=self.generic_name)

        if self.config_id in self.all_results:
            return

        logger.debug(f'****** Estimate {self.config_id}')
        the_biogeme = bio.BIOGEME.from_configuration(
            config_id=self.config_id, expression=self.expression, database=self.database
        )
        the_biogeme.modelName = self.model_names(self.config_id)
        logger.info(f'*** Estimate {the_biogeme.modelName}')
        the_biogeme.generate_html = False
        the_biogeme.generate_pickle = False
        results = the_biogeme.quickEstimate()
        self.all_results[self.config_id] = results

    def describe(self):
        """Short description of the solution. Used for reporting.

        :return: short description of the solution.
        :rtype: str
        """
        the_results = self.get_results()
        return f'{the_results.short_summary()}'

    def get_element(self, multi_objectives):
        """Obtains the element from the Pareto set corresponding to a specification

        :param multi_objectives: function calculating the objectives
            from the estimation results
        :type multi_objectives: fct(biogeme.results.bioResults) --> list[float]

        :return: element from the Pareto set
        :rtype: biogeme.pareto.SetElement

        """
        the_id = self.config_id
        the_results = self.get_results()
        the_objectives = multi_objectives(the_results)
        element = SetElement(the_id, the_objectives)
        logger.debug(f'{element=}')
        return element



""" Estimate the parameters of the nested logit models with the generated samples

:author: Michel Bierlaire
:date: Sat Apr 15 16:45:26 2023
"""

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta
from biogeme.sampling import mev_cnl_sampling

from sample import SAMPLE_SIZE as N_ALT
from choice_data import database

from model import V, log_probability

# Nests are defined with indicator variables. In this example, the nest
# "Asian" contains all the alternative i such as "Asian_i" is equal to 1.
# All other alternatives are alone in their own nest.
mu_nested = Beta('mu_nested', 1, 1, None, 0)
nests = {'Asian': mu_nested}

# In this context, the model is actually a cross-nested logit model,
# as the numbering is not consistent across observations. Indeed, if
# restaurant 2 is Asian for one observation, it may not be for the
# next.
log_gi = mev_cnl_sampling(V, None, log_probability, nests)
logprob = models.logmev(V, log_gi, None, 0)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'nested'

# Calculate the null log likelihood for reporting.
the_biogeme.calculateNullLoglikelihood({i: 1 for i in range(N_ALT)})

# Estimate the parameters
results = the_biogeme.estimate(recycle=True)
print(results.shortSummary())



"""File b05normal_mixture_simul.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:47:42 2023

Simulation of the mixture model, with estimation of the integration error.
"""

import sys
import numpy as np
import biogeme.biogeme as bio
from biogeme import models
import biogeme.results as res
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Beta, bioDraws, MonteCarlo

from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

try:
    import matplotlib.pyplot as plt

    PLOT = True
except ModuleNotFoundError:
    print('Install matplotlib to see the distribution of integration errors.')
    print('pip install matplotlib')
    PLOT = False

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation
B_TIME = Beta('B_TIME', 0, None, None, 0)

# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL')

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# The estimation results are read from the pickle file
try:
    results = res.bioResults(pickleFile='05normalMixture.pickle')
except BiogemeError:
    print(
        'Run first the script 05normalMixture.py in order to generate the '
        'file 05normalMixture.pickle.'
    )
    sys.exit()

# Conditional to B_TIME_RND, we have a logit model (called the kernel)
prob = models.logit(V, av, CHOICE)

# We calculate the integration error. Note that this formula assumes
# independent draws, and is not valid for Halton or antithetic draws.
integral = MonteCarlo(prob)
integralSquare = MonteCarlo(prob * prob)
variance = integralSquare - integral * integral
error = (variance / 2.0) ** 0.5

# And the value of the individual parameters
numerator = MonteCarlo(B_TIME_RND * prob)
denominator = integral

simulate = {
    'Numerator': numerator,
    'Denominator': denominator,
    'Integral': integral,
    'Integration error': error,
}

# Create the Biogeme object
biosim = bio.BIOGEME(database, simulate)
biosim.modelName = 'b05normal_mixture_simul'

# Simulate the requested quantities. The output is a Pandas data frame
simresults = biosim.simulate(results.getBetaValues())

# 95% confidence interval on the log likelihood.
simresults['left'] = np.log(
    simresults['Integral'] - 1.96 * simresults['Integration error']
)
simresults['right'] = np.log(
    simresults['Integral'] + 1.96 * simresults['Integration error']
)

print(f'Log likelihood: {np.log(simresults["Integral"]).sum()}')
print(
    f'Integration error for {biosim.number_of_draws} draws: '
    f'{simresults["Integration error"].sum()}'
)
print(f'In average {simresults["Integration error"].mean()} per observation.')
print(
    f'95% confidence interval: [{simresults["left"].sum()}-'
    f'{simresults["right"].sum()}]'
)

# Post processing to obtain the individual parameters
simresults['beta'] = simresults['Numerator'] / simresults['Denominator']

# Plot the histogram of individual parameters
if PLOT:
    simresults['beta'].plot(kind='hist', density=True, bins=20)


# Plot the general distribution of beta
def normalpdf(val, mu=0.0, std=1.0):
    """
    Calculate the pdf of the normal distribution, for plotting purposes.

    """
    d = -(val - mu) * (val - mu)
    n = 2.0 * std * std
    a = d / n
    num = np.exp(a)
    den = std * 2.506628275
    p = num / den
    return p


betas = results.getBetaValues(['B_TIME', 'B_TIME_S'])
x = np.arange(simresults['beta'].min(), simresults['beta'].max(), 0.01)
if PLOT:
    plt.plot(x, normalpdf(x, betas['B_TIME'], betas['B_TIME_S']), '-')
    plt.show()



"""
Test the segmentation module

:author: Michel Bierlaire
:date: Fri Dec 31 11:01:24 2021

"""

import unittest
from biogeme.expressions import Beta, Variable
import biogeme.segmentation as seg
import biogeme.exceptions as excep


class test_segmentation(unittest.TestCase):
    def test_one_segmentation(self):
        x = Variable('x')
        mapping = {1: '1st', 2: '2nd', 3: '3rd'}

        first_tuple = seg.DiscreteSegmentationTuple(variable=x, mapping=mapping)
        second_tuple = seg.DiscreteSegmentationTuple(
            variable=x, mapping=mapping, reference='3rd'
        )
        with self.assertRaises(excep.BiogemeError):
            third_tuple = seg.DiscreteSegmentationTuple(
                variable=x, mapping=mapping, reference='anything'
            )

        beta = Beta('beta', 1, None, None, 0)

        one_segmentation = seg.OneSegmentation(beta, first_tuple)

        beta_name = one_segmentation.beta_name('2nd')
        self.assertEqual(beta_name, 'beta_2nd')

        beta_expression = one_segmentation.beta_expression('2nd')
        self.assertEqual(str(beta_expression), 'beta_2nd(init=1)')

        beta_code = one_segmentation.beta_code('2nd', assignment=False)
        expected_code = "Beta('beta_2nd', 1, None, None, 0)"
        self.assertEqual(beta_code, expected_code)

        beta_code_assignment = one_segmentation.beta_code('2nd', assignment=True)
        expected_code_assignment = "beta_2nd = Beta('beta_2nd', 1, None, None, 0)"
        self.assertEqual(beta_code_assignment, expected_code_assignment)

        list_of_expressions = [str(e) for e in one_segmentation.list_of_expressions()]
        expected_list = [
            '(beta_2nd(init=1) * (x == `2.0`))',
            '(beta_3rd(init=1) * (x == `3.0`))',
        ]
        self.assertCountEqual(list_of_expressions, expected_list)

        list_of_code = one_segmentation.list_of_code()
        expected_list = [
            "beta_2nd * (Variable('x') == 2)",
            "beta_3rd * (Variable('x') == 3)",
        ]
        self.assertCountEqual(list_of_code, expected_list)

    def test_segmentation(self):
        variable_1 = Variable('Variable1')
        variable_2 = Variable('Variable2')
        parameter = Beta('test', 0, -10, 10, 0)
        mapping_1 = {
            1: 'var1',
            2: 'var2',
            3: 'var3',
            4: 'var4',
            5: 'var5',
        }
        mapping_2 = {
            10: 'my1',
            20: 'my2',
            30: 'my3',
            40: 'my4',
            50: 'my5',
        }
        first_segmentation = seg.DiscreteSegmentationTuple(
            variable=variable_1, mapping=mapping_1, reference='var1'
        )
        second_segmentation = seg.DiscreteSegmentationTuple(
            variable=variable_2, mapping=mapping_2, reference='my1'
        )

        the_segmentation = seg.Segmentation(
            parameter, (first_segmentation, second_segmentation)
        )

        beta_code = the_segmentation.beta_code()
        expected_code = "Beta('test', 0, -10, 10, 0)"
        self.assertEqual(beta_code, expected_code)

        segmented_expression = the_segmentation.segmented_beta()
        result_segment = (
            'bioMultSum('
            'test(init=0), '
            '(test_var2(init=0) * (Variable1 == `2.0`)), '
            '(test_var3(init=0) * (Variable1 == `3.0`)), '
            '(test_var4(init=0) * (Variable1 == `4.0`)), '
            '(test_var5(init=0) * (Variable1 == `5.0`)), '
            '(test_my2(init=0) * (Variable2 == `20.0`)), '
            '(test_my3(init=0) * (Variable2 == `30.0`)), '
            '(test_my4(init=0) * (Variable2 == `40.0`)), '
            '(test_my5(init=0) * (Variable2 == `50.0`)))'
        )

        self.assertEqual(str(segmented_expression), result_segment)
        segmented_code = the_segmentation.segmented_code()
        result_code = (
            "test_var2 = Beta('test_var2', 0, None, None, 0)\n"
            "test_var3 = Beta('test_var3', 0, None, None, 0)\n"
            "test_var4 = Beta('test_var4', 0, None, None, 0)\n"
            "test_var5 = Beta('test_var5', 0, None, None, 0)\n"
            "test_my2 = Beta('test_my2', 0, None, None, 0)\n"
            "test_my3 = Beta('test_my3', 0, None, None, 0)\n"
            "test_my4 = Beta('test_my4', 0, None, None, 0)\n"
            "test_my5 = Beta('test_my5', 0, None, None, 0)\n"
            "segmented_test = bioMultSum(["
            "Beta('test', 0, -10, 10, 0), "
            "test_var2 * (Variable('Variable1') == 2), "
            "test_var3 * (Variable('Variable1') == 3), "
            "test_var4 * (Variable('Variable1') == 4), "
            "test_var5 * (Variable('Variable1') == 5), "
            "test_my2 * (Variable('Variable2') == 20), "
            "test_my3 * (Variable('Variable2') == 30), "
            "test_my4 * (Variable('Variable2') == 40), "
            "test_my5 * (Variable('Variable2') == 50)])"
        )
        self.assertEqual(str(segmented_code), result_code)


if __name__ == '__main__':
    unittest.main()



""" Generation of various types of draws.

:author: Michel Bierlaire

:date: Tue Jun 18 19:05:13 2019
"""

# Too constraining
# pylint: disable=invalid-name, too-many-arguments, too-many-locals, too-many-statements

import numpy as np
import biogeme.exceptions as excep


def getUniform(sample_size, number_of_draws, symmetric=False):
    """Uniform [0, 1] or [-1, 1] numbers

    :param sample_size: number of observations for which draws must be
                       generated. If None, a one dimensional array
                       will be generated. If it has a values k, then k
                       series of draws will be generated
    :type sample_size: int
    :param number_of_draws: number of draws to generate.
    :type number_of_draws: int
    :param symmetric: if True, draws from [-1: 1] are generated.
        If False, draws from [0: 1] are generated.  Default: False
    :type symmetric: bool
    :return: numpy array with the draws
    :rtype: numpy.array

    Example::

        draws = dr.getUniform(sample_size=3, number_of_draws=10, symmetric=False)
        array(
            [[0.13053817, 0.63892308, 0.55031567, 0.26347854, 0.16730932,
              0.77745367, 0.48283887, 0.84247501, 0.20550219, 0.02373537],
             [0.68935846, 0.03363595, 0.36006669, 0.26709364, 0.54907706,
              0.22492104, 0.2494399 , 0.17323209, 0.52370401, 0.54091257],
             [0.40310204, 0.89916711, 0.86065005, 0.94277699, 0.09077065,
              0.40107731, 0.22554722, 0.47693135, 0.14058265, 0.17397031]]
        )

        draws = dr.getUniform(sample_size=3, number_of_draws=10, symmetric=True)
        array(
            [[ 0.74403237, -0.27995692,  0.33997421, -0.89405035, -0.129761  ,
               0.86593325,  0.30657422,  0.82435619,  0.498482  ,  0.24561616],
             [-0.48239607, -0.29257815, -0.98342034,  0.68392813, -0.25379429,
               0.49359859, -0.26459883,  0.14569724, -0.68860467, -0.40903446],
             [ 0.93251627, -0.85166912,  0.58096917,  0.39289882, -0.65088635,
               0.40114744, -0.61327161,  0.08900539, -0.20985417,  0.67542226]]
        )

    :raise BiogemeError: if the number of draws is not positive.

    :raise BiogemeError: if the sample size is not positive.
    """
    if number_of_draws <= 0:
        raise excep.BiogemeError(f'Invalid number of draws: {number_of_draws}.')

    if sample_size <= 0:
        raise excep.BiogemeError(
            f'Invalid sample size: {sample_size} when generating draws.'
        )
    totalSize = number_of_draws * sample_size

    uniformNumbers = np.random.uniform(size=totalSize)
    if symmetric:
        uniformNumbers = 2.0 * uniformNumbers - 1.0

    uniformNumbers.shape = (sample_size, number_of_draws)
    return uniformNumbers


def getLatinHypercubeDraws(
    sample_size, number_of_draws, symmetric=False, uniformNumbers=None
):
    """Implementation of the Modified Latin Hypercube Sampling proposed
    by `Hess et al., (2006)`_.

    .. _`Hess et al., (2006)`:
       https://doi.org/10.1016/j.trb.2004.10.005

    :param sample_size: number of observations for which draws must be
                       generated. If None, a one dimensional array
                       will be generated. If it has a values k, then k
                       series of draws will be generated
    :type sample_size: int
    :param number_of_draws: number of draws to generate.
    :type number_of_draws: int
    :param symmetric: if True, draws from [-1: 1] are generated.
       If False, draws from [0: 1] are generated.  Default: False
    :type symmetric: bool
    :param uniformNumbers: numpy with uniformly distributed numbers.
       If None, the numpy uniform number generator is used.
    :type uniformNumbers: numpy.array

    :return: numpy array with the draws
    :rtype: numpy.array

    Example::

        latinHypercube = dr.getLatinHypercubeDraws(sample_size=3,
                                                   number_of_draws=10)
        array([[0.43362897, 0.5275741 , 0.09215663, 0.94056236, 0.34376868,
                0.87195551, 0.41495219, 0.71736691, 0.23198736, 0.145561  ],
               [0.30520544, 0.78082964, 0.83591146, 0.2733167 , 0.53890906,
                0.61607469, 0.00699715, 0.17179441, 0.7557228 , 0.39733102],
               [0.49676864, 0.67073483, 0.9788854 , 0.5726069 , 0.11894558,
                0.05515471, 0.2640275 , 0.82093696, 0.92034628, 0.64866597]])

    :raise BiogemeError: if the number of draws is not positive.

    :raise BiogemeError: if the sample size is not positive.

    :raise BiogemeError: if the number of uniform draws is inconsistent.
    """
    if number_of_draws <= 0:
        raise excep.BiogemeError(f'Invalid number of draws: {number_of_draws}.')

    if sample_size <= 0:
        raise excep.BiogemeError(
            f'Invalid sample size: {sample_size} when generating draws.'
        )
    totalSize = number_of_draws * sample_size

    if uniformNumbers is None:
        uniformNumbers = np.random.uniform(size=totalSize)
    else:
        if uniformNumbers.size != totalSize:
            errorMsg = (
                f'A total of {totalSize} uniform draws '
                f'must be provided, and not {uniformNumbers.size}.'
            )
            raise excep.BiogemeError(errorMsg)

    uniformNumbers.shape = (totalSize,)
    numbers = np.array(
        [(float(i) + uniformNumbers[i]) / float(totalSize) for i in range(totalSize)]
    )
    if symmetric:
        numbers = 2.0 * numbers - 1.0

    np.random.shuffle(numbers)
    numbers.shape = (sample_size, number_of_draws)
    return numbers


def getHaltonDraws(
    sample_size, number_of_draws, symmetric=False, base=2, skip=0, shuffled=False
):
    """Generate Halton draws.
    Implementation by Cristian Arteaga, University of Nevada Las Vegas,

    :param sample_size: number of observations for which draws must be
                       generated. If None, a one dimensional array
                       will be generated. If it has a values k, then k
                       series of draws will be generated
    :type sample_size: int

    :param number_of_draws: number of draws to generate.
    :type number_of_draws: int

    :param symmetric: if True, draws from [-1: 1] are generated.
           If False, draws from [0: 1] are generated.  Default: False
    :type symmetric: bool

    :param base: generate Halton draws for a given basis.
            Ideally, it should be a prime number. Default: 2.
    :type base: int

    :param skip: the number of  elements of the sequence to be discarded.
    :type skip: int

    :param shuffled: if True, each series is shuffled
    :type shuffled: bool

    :return: numpy array with the draws
    :rtype: numpy.array

    Example::

        halton = dr.getHaltonDraws(sample_size=2, number_of_draws=10, base=3)
        array([[0.33333333, 0.66666667, 0.11111111, 0.44444444, 0.77777778,
                0.22222222, 0.55555556, 0.88888889, 0.03703704, 0.37037037],
               [0.7037037 , 0.14814815, 0.48148148, 0.81481481, 0.25925926,
                0.59259259, 0.92592593, 0.07407407, 0.40740741, 0.74074074]])

    :raise BiogemeError: if the number of draws is not positive.

    :raise BiogemeError: if the sample size is not positive.
    """
    if number_of_draws <= 0:
        raise excep.BiogemeError(f'Invalid number of draws: {number_of_draws}.')

    if sample_size <= 0:
        raise excep.BiogemeError(
            f'Invalid sample size: {sample_size} when generating draws.'
        )
    length = number_of_draws * sample_size
    req_length = length + skip + 1
    numbers = np.empty(req_length)
    numbers[0] = 0
    numbers_idx = 1
    t = 1
    while numbers_idx < req_length:
        d = 1 / base**t
        numbers_size = numbers_idx
        i = 1
        while i < base and numbers_idx < req_length:
            max_numbers = min(req_length - numbers_idx, numbers_size)
            numbers[numbers_idx : numbers_idx + max_numbers] = (
                numbers[:max_numbers] + d * i
            )
            numbers_idx += max_numbers
            i += 1
        t += 1
    numbers = numbers[skip + 1 : length + skip + 1]

    if shuffled:
        np.random.shuffle(numbers)

    if symmetric:
        numbers = 2.0 * numbers - 1.0

    numbers.shape = (sample_size, number_of_draws)
    return numbers


def getAntithetic(unif, sample_size, number_of_draws):
    """Returns antithetic uniform draws

    :param unif: function taking two arguments (sample_size, number_of_draws)
                 and returning U[0, 1] draws
    :type unif: function

    :param sample_size: number of observations for which draws must be
                       generated. If None, a one dimensional array
                       will be generated. If it has a values k, then k
                       series of draws will be generated
    :type sample_size: int

    :param number_of_draws: number of draws to generate.
    :type number_of_draws: int

    :return: numpy array with the antithetic draws
    :rtype: numpy.array

    Example::

          draws = dr.getAntithetic(dr.getUniform,
                                   sample_size=3,
                                   number_of_draws=10)

          array([[0.48592363, 0.13648133, 0.35925946, 0.32431338, 0.32997936,
                  0.51407637, 0.86351867, 0.64074054, 0.67568662, 0.67002064],
                 [0.89261997, 0.0331808 , 0.30767182, 0.93433648, 0.17196124,
                  0.10738003, 0.9668192 , 0.69232818, 0.06566352, 0.82803876],
                 [0.81095587, 0.96171364, 0.40984817, 0.72177258, 0.16481096,
                  0.18904413, 0.03828636, 0.59015183, 0.27822742, 0.83518904]])

    """
    R = int(number_of_draws / 2.0)
    draws = unif(sample_size, R)
    return np.concatenate((draws, 1 - draws), axis=1)


def getNormalWichuraDraws(
    sample_size, number_of_draws, uniformNumbers=None, antithetic=False
):
    """Generate pseudo-random numbers from a normal distribution N(0, 1)

    It uses the Algorithm AS241 by `Wichura (1988)`_
    which produces the normal deviate z corresponding to a given lower
    tail area of p; z is accurate to about 1 part in :math:`10^{16}`.

    .. _`Wichura (1988)`:
       http://www.jstor.org/stable/2347330

    :param sample_size: number of observations for which draws must be
                       generated. If None, a one dimensional array
                       will be generated. If it has a values k, then k
                       series of draws will be generated
    :type sample_size: int

    :param number_of_draws: number of draws to generate.
    :type number_of_draws: int

    :param uniformNumbers: numpy with uniformly distributed numbers.
               If None, the numpy uniform number generator is used.
    :type uniformNumbers: numpy.array
    :param antithetic: if True, only half of the draws are
                       actually generated, and the series are completed
                       with their antithetic version.
    :type antithetic: bool

    :return: numpy array with the draws
    :rtype: numpy.array

    Example::

        draws = dr.getNormalWichuraDraws(sample_size=3, number_of_draws=10)
        array(
            [[ 0.52418458, -1.04344204, -2.11642482,  0.48257162, -2.67188279,
              -1.89993283,  0.28251041, -0.38424425,  1.53182226,  0.30651874],
             [-0.7937038 , -0.07884121, -0.91005616, -0.98855175,  1.09405753,
              -0.5997651 , -1.70785113,  1.57571384, -0.33208723, -1.03510102],
             [-0.13853654,  0.92595498, -0.80136586,  1.68454196,  0.9955927 ,
              -0.28615154,  2.10635541,  0.0436191 , -0.25417774,  0.01026933]]
        )

    :raise BiogemeError: if the number of draws is not positive.

    :raise BiogemeError: if the sample size is not positive.

    """
    if number_of_draws <= 0:
        raise excep.BiogemeError(f'Invalid number of draws: {number_of_draws}.')

    if antithetic:
        if number_of_draws % 2 != 0:
            errorMsg = (
                f'Please specify an even number of draws for '
                f'antithetic draws. Requested number of '
                f'{number_of_draws}.'
            )
            raise excep.BiogemeError(errorMsg)
        number_of_draws = int(number_of_draws / 2.0)

    if sample_size <= 0:
        raise excep.BiogemeError(
            f'Invalid sample size: {sample_size} when generating draws.'
        )
    totalSize = number_of_draws * sample_size

    split2 = 5.0e00
    const1 = 0.180625e00
    const2 = 1.6e00
    a0 = 3.3871328727963666080e00
    a1 = 1.3314166789178437745e02
    a2 = 1.9715909503065514427e03
    a3 = 1.3731693765509461125e04
    a4 = 4.5921953931549871457e04
    a5 = 6.7265770927008700853e04
    a6 = 3.3430575583588128105e04
    a7 = 2.5090809287301226727e03
    b1 = 4.2313330701600911252e01
    b2 = 6.8718700749205790830e02
    b3 = 5.3941960214247511077e03
    b4 = 2.1213794301586595867e04
    b5 = 3.9307895800092710610e04
    b6 = 2.8729085735721942674e04
    b7 = 5.2264952788528545610e03
    c0 = 1.42343711074968357734e00
    c1 = 4.63033784615654529590e00
    c2 = 5.76949722146069140550e00
    c3 = 3.64784832476320460504e00
    c4 = 1.27045825245236838258e00
    c5 = 2.41780725177450611770e-01
    c6 = 2.27238449892691845833e-02
    c7 = 7.74545014278341407640e-04
    d1 = 2.05319162663775882187e00
    d2 = 1.67638483018380384940e00
    d3 = 6.89767334985100004550e-01
    d4 = 1.48103976427480074590e-01
    d5 = 1.51986665636164571966e-02
    d6 = 5.47593808499534494600e-04
    d7 = 1.05075007164441684324e-09
    e0 = 6.65790464350110377720e00
    e1 = 5.46378491116411436990e00
    e2 = 1.78482653991729133580e00
    e3 = 2.96560571828504891230e-01
    e4 = 2.65321895265761230930e-02
    e5 = 1.24266094738807843860e-03
    e6 = 2.71155556874348757815e-05
    e7 = 2.01033439929228813265e-07
    f1 = 5.99832206555887937690e-01
    f2 = 1.36929880922735805310e-01
    f3 = 1.48753612908506148525e-02
    f4 = 7.86869131145613259100e-04
    f5 = 1.84631831751005468180e-05
    f6 = 1.42151175831644588870e-07
    f7 = 2.04426310338993978564e-15

    if uniformNumbers is None:
        uniformNumbers = np.random.uniform(size=totalSize)
    elif uniformNumbers.size != totalSize:
        errorMsg = (
            f'A total of {totalSize} uniform draws must be '
            f'provided, and not {uniformNumbers.size}.'
        )
        raise excep.BiogemeError(errorMsg)
    uniformNumbers.shape = (totalSize,)

    q = uniformNumbers - 0.5
    draws = np.zeros(uniformNumbers.shape)
    r = np.zeros(uniformNumbers.shape)
    cond1 = np.abs(uniformNumbers) <= 0.45
    r[cond1] = const1 - q[cond1] * q[cond1]
    draws[cond1] = (
        q[cond1]
        * (
            (
                (
                    (
                        (((a7 * r[cond1] + a6) * r[cond1] + a5) * r[cond1] + a4)
                        * r[cond1]
                        + a3
                    )
                    * r[cond1]
                    + a2
                )
                * r[cond1]
                + a1
            )
            * r[cond1]
            + a0
        )
        / (
            (
                (
                    (
                        (((b7 * r[cond1] + b6) * r[cond1] + b5) * r[cond1] + b4)
                        * r[cond1]
                        + b3
                    )
                    * r[cond1]
                    + b2
                )
                * r[cond1]
                + b1
            )
            * r[cond1]
            + 1
        )
    )
    cond2 = np.abs(uniformNumbers) > 0.45
    cond2a = np.logical_and(cond2, q < 0.0)
    cond2b = np.logical_and(cond2, q >= 0.0)
    r[cond2a] = uniformNumbers[cond2a]
    r[cond2b] = 1 - uniformNumbers[cond2b]
    cond2c = np.logical_and(cond2, r <= 0)
    cond2d = np.logical_and(cond2, r > 0)
    draws[cond2c] = 0.0
    r[cond2d] = np.sqrt(-np.log(r[cond2d]))
    cond2d_a = np.logical_and(cond2d, r <= split2)
    cond2d_b = np.logical_and(cond2d, r > split2)
    r[cond2d_a] = r[cond2d_a] - const2
    draws[cond2d_a] = (
        (
            (
                (
                    (((c7 * r[cond2d_a] + c6) * r[cond2d_a] + c5) * r[cond2d_a] + c4)
                    * r[cond2d_a]
                    + c3
                )
                * r[cond2d_a]
                + c2
            )
            * r[cond2d_a]
            + c1
        )
        * r[cond2d_a]
        + c0
    ) / (
        (
            (
                (
                    (((d7 * r[cond2d_a] + d6) * r[cond2d_a] + d5) * r[cond2d_a] + d4)
                    * r[cond2d_a]
                    + d3
                )
                * r[cond2d_a]
                + d2
            )
            * r[cond2d_a]
            + d1
        )
        * r[cond2d_a]
        + 1
    )
    r[cond2d_b] = r[cond2d_b] - split2
    draws[cond2d_b] = (
        (
            (
                (
                    (((e7 * r[cond2d_b] + e6) * r[cond2d_b] + e5) * r[cond2d_b] + e4)
                    * r[cond2d_b]
                    + e3
                )
                * r[cond2d_b]
                + e2
            )
            * r[cond2d_b]
            + e1
        )
        * r[cond2d_b]
        + e0
    ) / (
        (
            (
                (
                    (((f7 * r[cond2d_b] + f6) * r[cond2d_b] + f5) * r[cond2d_b] + f4)
                    * r[cond2d_b]
                    + f3
                )
                * r[cond2d_b]
                + f2
            )
            * r[cond2d_b]
            + f1
        )
        * r[cond2d_b]
        + 1
    )
    draws[cond2a] = -draws[cond2a]

    draws.shape = (sample_size, number_of_draws)

    if antithetic:
        draws = np.concatenate((draws, -draws), axis=1)

    return draws



""" Module in charge of functionalities related to the sampling of alternatives

:author: Michel Bierlaire
:date: Wed Sep  7 15:54:55 2022
"""

from collections import defaultdict
from typing import NamedTuple, Set
import numpy as np
import pandas as pd
from tqdm import tqdm
import biogeme.exceptions as excep
import biogeme.expressions as expr


class StratumTuple(NamedTuple):
    subset: Set[int]
    sample_size: int


LOG_PROBA_COL = '_log_proba'


def sample_alternatives(alternatives, id_column, partition, chosen=None):
    """Performing the sampling of alternatives

    :param alternatives: Pandas data frame containing all the
        alternatives as rows. One column must contain a unique ID
        identifying the alternatives. The other columns contain
        variables to include in the data file.
    :type alternatives: pandas.DataFrame

    :param id_column: name of the columns where the IDs of the
        alternatives are stored.
    :type id_column: str

    :param partition: each StratumTuple contains a set of IDs
        characterizing the subset, and the sample size, that is the
        number of alternatives to randomly draw from the subset.
    :type partition: tuple(StratumTuple)

    :param chosen: ID of the chosen alternative, that must be included
        in the choice set. If None, no alternative is added
        deterministically to the choice set.
    :type chosen: int

    :raise BiogemeError: if one alternative belongs to several subsets
        of the partition.

    :raise BiogemeError: if a set in the partition is empty.

    :raise BiogemeError: if the chosen alternative is unknown.

    :raise BiogemeError: if the requested sample size for a stratum if
        larger than the size of the stratum

    :raise BiogemeError: if some alternative do not appear in the partition

    """
    # Verify that we have a partition
    nbr_unique_elements = len(set.union(*[s.subset for s in partition]))
    total_nbr = sum(list(len(s.subset) for s in partition))
    if nbr_unique_elements != total_nbr:
        error_msg = (
            f'This is not a partition. There are {nbr_unique_elements} '
            f'unique elements, and the total size of the partition '
            f'is {total_nbr}. Some elements are therefore present '
            f'in more than one subset.'
        )
        raise excep.BiogemeError(error_msg)

    if nbr_unique_elements != alternatives.shape[0]:
        error_msg = (
            f'The partitions contain {nbr_unique_elements} alternatives '
            f'while there are {alternatives.shape[0]} in the database'
        )
        raise excep.BiogemeError(error_msg)
    # Verify that all requested alternatives appear in the database of alternatives
    for stratum in partition:
        for alt in stratum.subset:
            if alt not in alternatives[id_column]:
                error_msg = (
                    f'Alternative {alt} does not appear in the database of alternaitves'
                )
                raise excep.BiogemeError(error_msg)

    results = []

    for stratum in partition:
        n = len(stratum.subset)
        if n == 0:
            error_msg = 'A stratum is empty'
            raise excep.BiogemeError(error_msg)

        k = stratum.sample_size
        if k > n:
            error_msg = f'Cannot draw {k} elements in a stratum of size {n}'
            raise excep.BiogemeError(error_msg)

        logproba = np.log(k) - np.log(n)
        subset = alternatives[alternatives[id_column].isin(stratum.subset)]
        if chosen is not None and chosen in stratum.subset:
            chosen_alternative = alternatives[alternatives[id_column] == chosen].copy()
            if len(chosen_alternative) < 1:
                error_msg = f'Unknown alternative: {chosen}'
                raise excep.BiogemeError(error_msg)
            if len(chosen_alternative) > 1:
                error_msg = f'Duplicate alternative: {chosen}'
                raise excep.BiogemeError(error_msg)
            chosen_alternative[LOG_PROBA_COL] = logproba
            results.append(chosen_alternative)

            subset = subset.drop(
                subset[subset[id_column] == chosen].index, axis='index'
            )
            n -= 1
            k -= 1

        if k > 0:
            sample = subset.sample(n=k, replace=False, axis='index', ignore_index=True)
            sample[LOG_PROBA_COL] = logproba
            results.append(sample)

    return pd.concat(results, ignore_index=True)


def sampling_of_alternatives(
    partition,
    individuals,
    choice_column,
    alternatives,
    id_column,
    always_include_chosen=True,
):
    """Generation of databases with samples of alternatives

      :param partition: each StratumTuple contains a set of IDs
        characterizing the subset, and the sample size, that is the
        number of alternatives to randomly draw from the subset.
    :type partition: tuple(StratumTuple)

    :param individuals: Pandas data frame containing all the
        individuals as rows. One column must contain the choice of
        each individual.
    :type individuals: pandas.DataFrame

    :param choice_column: name of the column containing the choice of
        each individual.
    :type choice_column: str

    :param alternatives: Pandas data frame containing all the
        alternatives as rows. One column must contain a unique ID
        identifying the alternatives. The other columns contain
        variables to include in the data file.
    :type alternatives: pandas.DataFrame

    :param id_column: name of the column containing the Ids of the
        alternatives.
    :type id_column: str

    :param always_include_chosen: if True, the chosen alternative is
        always included in the choice set with label 0.
    :type always_include_chosen: bool

    :return: data frame containing the data ready for Biogeme.
    :rtype: pandas.DataFrame

    """
    for index_ind, the_individual_row in tqdm(
        individuals.iterrows(), total=individuals.shape[0]
    ):
        choice = the_individual_row[choice_column]
        the_alternatives = alternatives.copy(deep=True)
        if always_include_chosen:
            chosen = choice
        else:
            chosen = None

        sample = sample_alternatives(
            the_alternatives, id_column, partition, chosen=chosen
        )
        sample.reset_index(inplace=True, drop=True)

        if always_include_chosen:
            # Position the chosen alternative at the first row
            chosen_alternative = sample.index[sample[id_column] == choice].tolist()[0]
            new_index = [chosen_alternative] + [
                i for i in range(len(sample)) if i != chosen_alternative
            ]
            sample = sample.reindex(new_index).reset_index(drop=True)

        row_elements = {
            f'{c}_{index}': alt[c]
            for index, alt in sample.iterrows()
            for c in sample.columns
        }
        individuals.loc[index_ind, row_elements.keys()] = row_elements.values()

    return individuals


def mev_cnl_sampling(V, availability, sampling_log_probability, nests):
    """Generate the expression of the CNL G_i function in the context
    of sampling of alternatives.

    It is assumed that the following variables are available in the
        data: for each nest m and each alternative i, a variable m_i
        that is the level of membership of alternative i to nest m.

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param sampling_log_probability: if not None, it means that the
        choice set is actually a subset that has been sampled from the
        full choice set. In that case, this is a dictionary mapping
        each alternative with the logarithm of its probability to be
        selected in the sample.
    :type sampling_log_probability: dict(int: biogeme.expressions.Expression)

    :param nests: a dictionary where the keys are the names of the
        nests, and the values are the nest parameters.
    :type nests: dict(str: biogeme.expressions.Beta)

    """

    Gi_terms = defaultdict(list)
    biosum = {}
    for nest, mu in nests.items():
        if availability is None:
            biosum = expr.bioMultSum(
                [
                    expr.exp(sampling_log_probability[i])
                    * expr.Variable(f'{nest}_{i}') ** mu
                    * expr.exp(mu * util)
                    for i, util in V.items()
                ]
            )
        else:
            biosum = expr.bioMultSum(
                [
                    expr.exp(sampling_log_probability[i])
                    * availability[i]
                    * expr.Variable(f'{nest}_{i}') ** mu
                    * expr.exp(mu * util)
                    for i, util in V.items()
                ]
            )
        for i, util in V.items():
            Gi_terms[i].append(
                expr.Variable(f'{nest}_{i}') ** mu
                * expr.exp((mu - 1) * (V[i]))
                * biosum ** ((1.0 / mu) - 1.0)
            )
    log_gi = {
        k: expr.logzero(expr.bioMultSum(G)) if G else expr.Numeric(0)
        for k, G in Gi_terms.items()
    }
    log_gi = {
        k: G if G == 0 else G - sampling_log_probability[k] for k, G in log_gi.items()
    }
    return log_gi



"""File results_analysis

:author: Michel Bierlaire, EPFL
:date: Thu Jul 13 16:32:45 2023

Reports the results of the catalog estimation
"""


from biogeme.results import compile_estimation_results, pareto_optimal


def report(dict_of_results):
    """Reports the results of the estimared catalogs"""
    print(f'A total of {len(dict_of_results)} models have been estimated')
    print('== Estimation results ==')

    compiled_results, specs = compile_estimation_results(
        dict_of_results, use_short_names=True
    )
    print(compiled_results)
    for short_name, spec in specs.items():
        print(f'{short_name}\t{spec}')

    pareto_results = pareto_optimal(dict_of_results)
    compiled_pareto_results, pareto_specs = compile_estimation_results(
        pareto_results, use_short_names=True
    )
    print(compiled_pareto_results)
    for short_name, spec in pareto_specs.items():
        print(f'{short_name}\t{spec}')



"""File b06unif_mixture_MHLS.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:50:28 2023

 Example of a mixture of logit models, using Monte-Carlo integration.
 The mixing distribution is uniform.
 The draws are from the Modified Hypercube Latin Square
"""

import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import (
    Beta,
    bioDraws,
    exp,
    log,
    MonteCarlo,
)
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b06unif_mixture_MHLS')

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation
B_TIME = Beta('B_TIME', 0, None, None, 0)

# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)

# Define a random parameter, uniformly distributed, designed to be used
# for Monte-Carlo simulation
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL_MLHS')

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Conditional to B_TIME_RND, we have a logit model (called the kernel)
prob = exp(models.loglogit(V, av, CHOICE))
# We integrate over B_TIME_RND using Monte-Carlo
logprob = log(MonteCarlo(prob))

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = '06unif_mixture_MHLS'

# Estimate the parameters
results = the_biogeme.estimate()
print(results.short_summary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""File b05revenues.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 21:02:19 2023

We use an estimated model to calculate revenues
"""
import sys
import numpy as np
try: 
    import matplotlib.pyplot as plt
    can_plot = True
except ModuleNotFoundError:
    can_plot = False
from tqdm import tqdm
from biogeme import models
import biogeme.exceptions as excep
import biogeme.biogeme as bio
import biogeme.results as res
from optima_data import database, normalized_weight
from scenarios import scenario

# Read the estimation results from the file
try:
    results = res.bioResults(pickleFile='b02estimation.pickle')
except excep.BiogemeError:
    sys.exit(
        'Run first the script b02simulation.py '
        'in order to generate the '
        'file b02estimation.pickle.'
    )


def revenues(factor):
    """Calculate the total revenues generated by public transportation,
        when the price is multiplied by a factor.

    :param factor: factor that multiplies the current cost of public
        transportation
    :type factor: float

    :return: total revenues, followed by the lower and upper bound of
        the confidence interval.
    :rtype: tuple(float, float, float)

    """
    # Obtain the specification for the default scenario
    V, nests, _, marginal_cost_scenario = scenario(factor=factor)

    # Obtain the expression for the choice probability of each alternative
    prob_pt = models.nested(V, None, nests, 0)

    # We now simulate the choice probabilities,the weight and the
    # price variable

    simulate = {
        'weight': normalized_weight,
        'Revenue public transportation': prob_pt * marginal_cost_scenario,
    }

    the_biogeme = bio.BIOGEME(database, simulate)
    simulated_values = the_biogeme.simulate(results.getBetaValues())

    # We also calculate confidence intervals for the calculated quantities

    betas = the_biogeme.free_beta_names()
    beta_bootstrap = results.getBetasForSensitivityAnalysis(betas)
    left, right = the_biogeme.confidenceIntervals(beta_bootstrap, 0.9)

    revenues_pt = (
        simulated_values['Revenue public transportation'] * simulated_values['weight']
    ).sum()
    revenues_pt_left = (left['Revenue public transportation'] * left['weight']).sum()
    revenues_pt_right = (right['Revenue public transportation'] * right['weight']).sum()
    return revenues_pt, revenues_pt_left, revenues_pt_right


# Current revenues for public transportation

r, r_left, r_right = revenues(factor=1.0)
print(
    f'Total revenues for public transportation (for the sample): {r:.1f} CHF '
    f'[{r_left:.1f} CHF, '
    f'{r_right:.1f} CHF]'
)

# We now investigate how the revenues vary with the multiplicative factor

factors = np.arange(0.0, 5.0, 0.05)

plot_revenues = [revenues(s) for s in tqdm(factors)]
zipped = zip(*plot_revenues)
rev = next(zipped)
lower = next(zipped)
upper = next(zipped)

largest_revenue = max(rev)
max_index = rev.index(largest_revenue)
print(
    f'Largest revenue: {largest_revenue:.1f} obtained with '
    f'factor {factors[max_index]:.1f}'
)

if can_plot:
    # We plot the results
    ax = plt.gca()
    ax.plot(factors, rev, label="Revenues")
    ax.plot(factors, lower, label="Lower bound of the CI")
    ax.plot(factors, upper, label="Upper bound of the CI")
    ax.legend()

    plt.show()



"""Class that provides some automatic specification for segmented parameters

:author: Michel Bierlaire
:date: Thu Feb  2 09:42:36 2023

"""

from biogeme.expressions import Beta, bioMultSum, Variable, Numeric
import biogeme.exceptions as excep


class DiscreteSegmentationTuple:
    """Characterization of a segmentation"""

    def __init__(self, variable, mapping, reference=None):
        """Ctor

        :param variable: socio-economic variable used for the segmentation, or its name
        :type variable: biogeme.expressions.Variable or str

        :param mapping: maps the values of the variable with the name of a category
        :type mapping: dict(int: str)

        :param reference: name of the reference category. If None, an
            arbitrary category is selected as reference.  :type:
        :type reference: str

        :raise BiogemeError: if the name of the reference category
            does not appear in the list.

        """

        self.variable = (
            variable if isinstance(variable, Variable) else Variable(variable)
        )
        self.mapping = mapping
        if reference is None:
            self.reference = next(iter(mapping.values()))
        elif reference not in mapping.values():
            error_msg = (
                f'Reference category {reference} does not appear in the list '
                f'of categories: {mapping.values()}'
            )
            raise excep.BiogemeError(error_msg)
        else:
            self.reference = reference

        if self.reference is None:
            raise excep.BiogemeError('Reference should not be None')

    def __repr__(self):
        result = f'{self.variable.name}: [{self.mapping}] ref: {self.reference}'
        return result

    def __str__(self):
        result = f'{self.variable.name}: [{self.mapping}] ref: {self.reference}'
        return result


class OneSegmentation:
    """Single segmentation of a parameter"""

    def __init__(self, beta, segmentation_tuple):
        """Ctor

        :param beta: parameter to be segmented
        :type beta: biogeme.expressions.Beta

        :param segmentation_tuple: characterization of the segmentation
        :type segmentation_tuple: DiscreteSegmentationTuple
        """
        self.beta = beta
        self.variable = segmentation_tuple.variable
        self.reference = segmentation_tuple.reference
        self.mapping = {
            k: v for k, v in segmentation_tuple.mapping.items() if v != self.reference
        }

    def beta_name(self, category):
        """Construct the name of the parameter associated with a specific category

        :param category: name of the category
        :type category: str

        :return: name of parameter for the category
        :rtype: str

        :raise BiogemeError: if the category is not listed in the
            mapping of the segmentation.
        """
        if category not in self.mapping.values():
            error_msg = (
                f'Unknown category: {category}. List of known categories: '
                f'{self.mapping.values()}'
            )
            raise excep.BiogemeError(error_msg)
        return f'{self.beta.name}_{category}'

    def beta_expression(self, category):
        """Constructs the expression for the parameter associated with
            a specific category

        :param category: name of the category
        :type category: str

        :return: expression of the parameter for the category
        :rtype: biogeme.expressions.Beta

        """
        name = self.beta_name(category)
        if category == self.reference:
            lower_bound = self.beta.lb
            upper_bound = self.beta.ub
        else:
            lower_bound = None
            upper_bound = None

        return Beta(
            name,
            self.beta.initValue,
            lower_bound,
            upper_bound,
            self.beta.status,
        )

    def beta_code(self, category, assignment):
        """Constructs the Python code for the expression of the
            parameter associated with a specific category

        :param category: name of the category
        :type category: str

        :param assignment: if True, the code includes the assigbnment to a variable.
        :type assignment: bool

        :return: the Python code
        :rtype: str
        """
        if category == self.reference:
            lower_bound = self.beta.lb
            upper_bound = self.beta.ub
        else:
            lower_bound = None
            upper_bound = None
        name = self.beta_name(category)
        if assignment:
            return (
                f"{name} = Beta('{name}', {self.beta.initValue}, "
                f"{lower_bound}, {upper_bound}, {self.beta.status})"
            )
        return (
            f"Beta('{name}', {self.beta.initValue}, {lower_bound}, "
            f"{upper_bound}, {self.beta.status})"
        )

    def list_of_expressions(self):
        """Create a list of expressions involved in the segmentation of the parameter

        :return: list of expressions
        :rtype: list(biogeme.expressions.Expression)

        """
        terms = [
            self.beta_expression(category) * (self.variable == Numeric(value))
            for value, category in self.mapping.items()
        ]
        return terms

    def list_of_code(self):
        """Create a list of Python codes for the expressions involved
            in the segmentation of the parameter

        :return: list of codes
        :rtype: list(str)

        """
        return [
            (
                f"{self.beta_name(category)} "
                f"* (Variable('{self.variable.name}') == {value})"
            )
            for value, category in self.mapping.items()
        ]


class Segmentation:
    """Segmentation of a parameter, possibly with multiple socio-economic variables"""

    def __init__(self, beta, segmentation_tuples, prefix='segmented'):
        """Ctor

        :param beta: parameter to be segmented
        :type beta: biogeme.expressions.Beta

        :param segmentation_tuples: characterization of the segmentations
        :type segmentation_tuples: list(DiscreteSegmentationTuple)

        :param prefix: prefix to be used to generated the name of the
            segmented parameter
        :type prefix: str
        """
        self.beta = beta
        self.segmentations = tuple(
            OneSegmentation(beta, s) for s in segmentation_tuples
        )
        self.prefix = prefix

    def beta_code(self):
        """Constructs the Python code for the parameter

        :return: Python code
        :rtype: str
        """

        beta_name = f"'{self.beta.name}'"
        return (
            f'Beta({beta_name}, {self.beta.initValue}, {self.beta.lb}, '
            f'{self.beta.ub}, {self.beta.status})'
        )

    def segmented_beta(self):
        """Create an expressions that combines all the segments

        :return: combined expression
        :rtype: biogeme.expressions.Expression

        """
        ref_beta = Beta(
            name=self.beta.name,
            value=self.beta.initValue,
            lowerbound=self.beta.lb,
            upperbound=self.beta.ub,
            status=self.beta.status,
        )
        terms = [ref_beta]
        terms += [
            element for s in self.segmentations for element in s.list_of_expressions()
        ]

        return bioMultSum(terms)

    def segmented_code(self):
        """Create the Python code for an expressions that combines all the segments

        :return: Python code for the combined expression
        :rtype: str
        """
        result = '\n'.join(
            [
                s.beta_code(c, assignment=True)
                for s in self.segmentations
                for c in s.mapping.values()
            ]
        )
        result += '\n'

        terms = [self.beta_code()]
        terms += [element for s in self.segmentations for element in s.list_of_code()]

        if len(terms) == 1:
            result += terms[0]
        else:
            joined_terms = ', '.join(terms)
            result += f'{self.prefix}_{self.beta.name} = bioMultSum([{joined_terms}])'
        return result


def segmented_beta(beta, segmentation_tuples, prefix='segmented'):
    """Obtain the segmented beta from a unique function call

    :param beta: parameter to be segmented
    :type beta: biogeme.expressions.Beta

    :param segmentation_tuples: characterization of the segmentations
    :type segmentation_tuples: list(DiscreteSegmentationTuple)

    :param prefix: prefix to be used to generated the name of the
        segmented parameter
    :type prefix: str

    :return: expression of the segmented beta
    :rtype: biogeme.expressions.Expression
    """
    the_segmentation = Segmentation(
        beta=beta, segmentation_tuples=segmentation_tuples, prefix=prefix
    )
    return the_segmentation.segmented_beta()



"""File b15panel_discrete_bis.py

:author: Michel Bierlaire, EPFL
:date: Mon Apr 10 11:55:26 2023

 Example of a discrete mixture of logit models, also called latent class model.
 The datafile is organized as panel data.
 Here, we integrate before the discrete mixture to show that it is equivalent.
"""

import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import (
    Beta,
    bioDraws,
    PanelLikelihoodTrajectory,
    MonteCarlo,
    log,
    exp,
)
from swissmetro_panel import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b15panel_discrete_bis.py')

# Parameters to be estimated. One version for each latent class.
NUMBER_OF_CLASSES = 2
B_COST = [Beta(f'B_COST_class{i}', 0, None, None, 0) for i in range(NUMBER_OF_CLASSES)]

# Define a random parameter, normally distributed across individuals,
# designed to be used for Monte-Carlo simulation
B_TIME = [Beta(f'B_TIME_class{i}', 0, None, None, 0) for i in range(NUMBER_OF_CLASSES)]

# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = [
    Beta(f'B_TIME_S_class{i}', 1, None, None, 0) for i in range(NUMBER_OF_CLASSES)
]
B_TIME_RND = [
    B_TIME[i] + B_TIME_S[i] * bioDraws(f'B_TIME_RND_class{i}', 'NORMAL_ANTI')
    for i in range(NUMBER_OF_CLASSES)
]

# We do the same for the constants, to address serial correlation.
ASC_CAR = [
    Beta(f'ASC_CAR_class{i}', 0, None, None, 0) for i in range(NUMBER_OF_CLASSES)
]
ASC_CAR_S = [
    Beta(f'ASC_CAR_S_class{i}', 1, None, None, 0) for i in range(NUMBER_OF_CLASSES)
]
ASC_CAR_RND = [
    ASC_CAR[i] + ASC_CAR_S[i] * bioDraws(f'ASC_CAR_RND_class{i}', 'NORMAL_ANTI')
    for i in range(NUMBER_OF_CLASSES)
]

ASC_TRAIN = [
    Beta(f'ASC_TRAIN_class{i}', 0, None, None, 0) for i in range(NUMBER_OF_CLASSES)
]
ASC_TRAIN_S = [
    Beta(f'ASC_TRAIN_S_class{i}', 1, None, None, 0) for i in range(NUMBER_OF_CLASSES)
]
ASC_TRAIN_RND = [
    ASC_TRAIN[i] + ASC_TRAIN_S[i] * bioDraws(f'ASC_TRAIN_RND_class{i}', 'NORMAL_ANTI')
    for i in range(NUMBER_OF_CLASSES)
]

ASC_SM = [Beta(f'ASC_SM_class{i}', 0, None, None, 1) for i in range(NUMBER_OF_CLASSES)]
ASC_SM_S = [
    Beta(f'ASC_SM_S_class{i}', 1, None, None, 0) for i in range(NUMBER_OF_CLASSES)
]
ASC_SM_RND = [
    ASC_SM[i] + ASC_SM_S[i] * bioDraws(f'ASC_SM_RND_class{i}', 'NORMAL_ANTI')
    for i in range(NUMBER_OF_CLASSES)
]

# Class memebership probability
PROB_class0 = exp(Beta('PROB_class0', 0.5, 0, 1, 0)) / (
    1 + exp(Beta('PROB_class0', 0, None, None, 0))
)
PROB_class1 = 1 - PROB_class0

# In class 0, it is assumed that the time coefficient is zero
B_TIME_RND[0] = 0

# Utility functions
V1 = [
    ASC_TRAIN_RND[i] + B_TIME_RND[i] * TRAIN_TT_SCALED + B_COST[i] * TRAIN_COST_SCALED
    for i in range(NUMBER_OF_CLASSES)
]
V2 = [
    ASC_SM_RND[i] + B_TIME_RND[i] * SM_TT_SCALED + B_COST[i] * SM_COST_SCALED
    for i in range(NUMBER_OF_CLASSES)
]
V3 = [
    ASC_CAR_RND[i] + B_TIME_RND[i] * CAR_TT_SCALED + B_COST[i] * CAR_CO_SCALED
    for i in range(NUMBER_OF_CLASSES)
]
V = [{1: V1[i], 2: V2[i], 3: V3[i]} for i in range(NUMBER_OF_CLASSES)]

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# The choice model is a discrete mixture of logit, with availability conditions
# We calculate the conditional probability for each class
prob = [
    MonteCarlo(PanelLikelihoodTrajectory(models.logit(V[i], av, CHOICE)))
    for i in range(NUMBER_OF_CLASSES)
]

# Conditional to the random variables, likelihood for the individual.
probIndiv = PROB_class0 * prob[0] + PROB_class1 * prob[1]

# We integrate over the random variables using Monte-Carlo
logprob = log(probIndiv)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b15panel_discrete_bis'

# Estimate the parameters.
results = the_biogeme.estimate()
print(results.short_summary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""Implementation of the class Database, wrapping a pandas dataframe
for specific services to Biogeme

:author: Michel Bierlaire

:date: Tue Mar 26 16:42:54 2019

"""

import logging
from typing import NamedTuple
import numpy as np
import pandas as pd

from biogeme.segmentation import DiscreteSegmentationTuple
import biogeme.exceptions as excep
import biogeme.filenames as bf
from biogeme import tools
from biogeme import draws

from biogeme.expressions import Variable, isNumeric, Numeric


class EstimationValidation(NamedTuple):
    estimation: pd.DataFrame
    validation: pd.DataFrame


logger = logging.getLogger(__name__)
"""Logger that controls the output of
        messages to the screen and log file.
        """


class Database:
    """Class that contains and prepare the database."""

    # @staticmethod
    def uniform_antithetic(sample_size, number_of_draws):
        return draws.getAntithetic(draws.getUniform, sample_size, number_of_draws)

    # @staticmethod
    def halton2(sample_size, number_of_draws):
        return draws.getHaltonDraws(sample_size, number_of_draws, base=2, skip=10)

    # @staticmethod
    def halton3(sample_size, number_of_draws):
        return draws.getHaltonDraws(sample_size, number_of_draws, base=3, skip=10)

    # @staticmethod
    def halton5(sample_size, number_of_draws):
        return draws.getHaltonDraws(sample_size, number_of_draws, base=5, skip=10)

    # @staticmethod
    def MLHS_anti(sample_size, number_of_draws):
        return draws.getAntithetic(
            draws.getLatinHypercubeDraws, sample_size, number_of_draws
        )

    # @staticmethod
    def symm_uniform(sample_size, number_of_draws):
        return draws.getUniform(sample_size, number_of_draws, symmetric=True)

    # @staticmethod
    def symm_uniform_antithetic(sample_size, number_of_draws):
        R = int(number_of_draws / 2)
        localDraws = Database.symm_uniform(sample_size, R)
        return np.concatenate((localDraws, -localDraws), axis=1)

    # @staticmethod
    def symm_halton2(sample_size, number_of_draws):
        return draws.getHaltonDraws(
            sample_size, number_of_draws, symmetric=True, base=2, skip=10
        )

    # @staticmethod
    def symm_halton3(sample_size, number_of_draws):
        return draws.getHaltonDraws(
            sample_size, number_of_draws, symmetric=True, base=3, skip=10
        )

    # @staticmethod
    def symm_halton5(sample_size, number_of_draws):
        return draws.getHaltonDraws(
            sample_size, number_of_draws, symmetric=True, base=5, skip=10
        )

    # @staticmethod
    def symm_MLHS(sample_size, number_of_draws):
        return draws.getLatinHypercubeDraws(
            sample_size, number_of_draws, symmetric=True
        )

    # @staticmethod
    def symm_MLHS_anti(sample_size, number_of_draws):
        R = int(number_of_draws / 2)
        localDraws = Database.symm_MLHS(sample_size, R)
        return np.concatenate((localDraws, -localDraws), axis=1)

    # @staticmethod
    def normal_antithetic(sample_size, number_of_draws):
        return draws.getNormalWichuraDraws(
            sample_size=sample_size,
            number_of_draws=number_of_draws,
            antithetic=True,
        )

    # @staticmethod
    def normal_halton2(sample_size, number_of_draws):
        unif = draws.getHaltonDraws(sample_size, number_of_draws, base=2, skip=10)
        return draws.getNormalWichuraDraws(
            sample_size,
            number_of_draws,
            uniformNumbers=unif,
            antithetic=False,
        )

    # @staticmethod
    def normal_halton3(sample_size, number_of_draws):
        unif = draws.getHaltonDraws(sample_size, number_of_draws, base=2, skip=10)
        return draws.getNormalWichuraDraws(
            sample_size,
            number_of_draws,
            uniformNumbers=unif,
            antithetic=False,
        )

    # @staticmethod
    def normal_halton5(sample_size, number_of_draws):
        unif = draws.getHaltonDraws(sample_size, number_of_draws, base=2, skip=10)
        return draws.getNormalWichuraDraws(
            sample_size,
            number_of_draws,
            uniformNumbers=unif,
            antithetic=False,
        )

    # @staticmethod
    def normal_MLHS(sample_size, number_of_draws):
        unif = draws.getLatinHypercubeDraws(sample_size, number_of_draws)
        return draws.getNormalWichuraDraws(
            sample_size,
            number_of_draws,
            uniformNumbers=unif,
            antithetic=False,
        )

    # @staticmethod
    def normal_MLHS_anti(sample_size, number_of_draws):
        unif = draws.getLatinHypercubeDraws(sample_size, int(number_of_draws / 2.0))
        return draws.getNormalWichuraDraws(
            sample_size, number_of_draws, uniformNumbers=unif, antithetic=True
        )

    # Dictionary containing native random number generators. Class attribute
    nativeRandomNumberGenerators = {
        'UNIFORM': (draws.getUniform, 'Uniform U[0, 1]'),
        'UNIFORM_ANTI': (uniform_antithetic, 'Antithetic uniform U[0, 1]'),
        'UNIFORM_HALTON2': (
            halton2,
            'Halton draws with base 2, skipping the first 10',
        ),
        'UNIFORM_HALTON3': (
            halton3,
            'Halton draws with base 3, skipping the first 10',
        ),
        'UNIFORM_HALTON5': (
            halton5,
            'Halton draws with base 5, skipping the first 10',
        ),
        'UNIFORM_MLHS': (
            draws.getLatinHypercubeDraws,
            'Modified Latin Hypercube Sampling on [0, 1]',
        ),
        'UNIFORM_MLHS_ANTI': (
            MLHS_anti,
            'Antithetic Modified Latin Hypercube Sampling on [0, 1]',
        ),
        'UNIFORMSYM': (symm_uniform, 'Uniform U[-1, 1]'),
        'UNIFORMSYM_ANTI': (
            symm_uniform_antithetic,
            'Antithetic uniform U[-1, 1]',
        ),
        'UNIFORMSYM_HALTON2': (
            symm_halton2,
            'Halton draws on [-1, 1] with base 2, skipping the first 10',
        ),
        'UNIFORMSYM_HALTON3': (
            symm_halton3,
            'Halton draws on [-1, 1] with base 3, skipping the first 10',
        ),
        'UNIFORMSYM_HALTON5': (
            symm_halton5,
            'Halton draws on [-1, 1] with base 5, skipping the first 10',
        ),
        'UNIFORMSYM_MLHS': (
            symm_MLHS,
            'Modified Latin Hypercube Sampling on [-1, 1]',
        ),
        'UNIFORMSYM_MLHS_ANTI': (
            symm_MLHS_anti,
            'Antithetic Modified Latin Hypercube Sampling on [-1, 1]',
        ),
        'NORMAL': (draws.getNormalWichuraDraws, 'Normal N(0, 1) draws'),
        'NORMAL_ANTI': (normal_antithetic, 'Antithetic normal draws'),
        'NORMAL_HALTON2': (
            normal_halton2,
            'Normal draws from Halton base 2 sequence',
        ),
        'NORMAL_HALTON3': (
            normal_halton3,
            'Normal draws from Halton base 3 sequence',
        ),
        'NORMAL_HALTON5': (
            normal_halton5,
            'Normal draws from Halton base 5 sequence',
        ),
        'NORMAL_MLHS': (
            normal_MLHS,
            'Normal draws from Modified Latin Hypercube Sampling',
        ),
        'NORMAL_MLHS_ANTI': (
            normal_MLHS_anti,
            ('Antithetic normal draws from Modified Latin Hypercube Sampling'),
        ),
    }

    # This statement does not work for versions of python before 3.10
    # @staticmethod
    def descriptionOfNativeDraws():
        """Describe the draws available draws with Biogeme

        :return: dict, where the keys are the names of the draws,
                 and the value their description

        Example of output::

            {'UNIFORM: Uniform U[0, 1]',
             'UNIFORM_ANTI: Antithetic uniform U[0, 1]'],
             'NORMAL: Normal N(0, 1) draws'}

        :rtype: dict

        """
        return {
            key: tuple[1]
            for key, tuple in Database.nativeRandomNumberGenerators.items()
        }

    def __init__(self, name, pandasDatabase):
        """Constructor

        :param name: name of the database.
        :type name: string

        :param pandasDatabase: data stored in a pandas data frame.
        :type pandasDatabase: pandas.DataFrame

        :raise BiogemeError: if the audit function detects errors.
        :raise BiogemeError: if the database is empty.
        """

        self.name = name
        """ Name of the database. Used mainly for the file name when
        dumping data.
        """

        if len(pandasDatabase.index) == 0:
            error_msg = 'Database has no entry'
            raise excep.BiogemeError(error_msg)

        self.data = pandasDatabase  #: Pandas data frame containing the data.

        self.fullData = pandasDatabase
        """Pandas data frame containing the full data. Useful when batches of
        the sample are used for approximating the log likelihood.
        """

        self.variables = None
        """names of the headers of the database so that they can be used as
        an object of type biogeme.expressions.Expression. Initialized
        by _generateHeaders()
        """

        self._generateHeaders()

        self.excludedData = 0
        """Number of observations removed by the function
        :meth:`biogeme.Database.remove`
        """

        self.panelColumn = None
        """Name of the column identifying the individuals in a panel
        data context. None if data is not panel.
        """

        self.individualMap = None
        """map identifying the range of observations for each individual in a
        panel data context. None if data is not panel.
        """

        self.fullIndividualMap = None
        """complete map identifying the range of observations for each
        individual in a panel data context. None if data is not
        panel. Useful when batches of the sample are used to
        approximate the log likelihood function.
        """

        self.userRandomNumberGenerators = {}
        """Dictionary containing user defined random number
        generators. Defined by the function
        Database.setRandomNumberGenerators that checks that reserved
        keywords are not used. The element of the dictionary is a
        tuple with two elements: (0) the function generating the
        draws, and (1) a string describing the type of draws
        """

        self.number_of_draws = 0
        """Number of draws generated by the function Database.generateDraws.
        Value 0 if this function is not called.
        """

        self.typesOfDraws = {}  #: Types of draws for Monte Carlo integration

        self.theDraws = None  #: Draws for Monte-Carlo integration

        self._avail = None  #: Availability expression to check

        self._choice = None  #: Choice expression to check

        self._expression = None  #: Expression to check

        listOfErrors, _ = self._audit()
        # For now, the audit issues only errors. If warnings are
        # triggered in the future, the nexrt lines should be
        # uncommented.
        # if listOfWarnings:
        #    logger.warning('\n'.join(listOfWarnings))
        if listOfErrors:
            logger.warning('\n'.join(listOfErrors))
            raise excep.BiogemeError('\n'.join(listOfErrors))

    def _audit(self):
        """Performs a series of checks and reports warnings and errors.
          - Check if there are non numerical entries.
          - Check if there are NaN (not a number) entries.
          - Check if there are strings.
          - Check if the numbering of individuals are contiguous
            (panel data only).

        :return: A tuple of two lists with the results of the diagnostic:
            listOfErrors, listOfWarnings
        :rtype: tuple(list(str), list(str))
        """
        listOfErrors = []
        listOfWarnings = []
        for col, dtype in self.data.dtypes.items():
            if not np.issubdtype(dtype, np.number):
                theError = f'Column {col} in the database does contain {dtype}'
                listOfErrors.append(theError)

        if self.data.isnull().values.any():
            theError = (
                'The database contains NaN value(s). '
                'Detect where they are using the function isnan()'
            )
            listOfErrors.append(theError)

        return listOfErrors, listOfWarnings

    def _generateHeaders(self):
        """Record the names of the headers
        of the database so that they can be used as an object of type
        biogeme.expressions.Expression
        """
        self.variables = {col: Variable(col) for col in self.data.columns}

    def valuesFromDatabase(self, expression):
        """Evaluates an expression for each entry of the database.

        :param expression: expression to evaluate
        :type expression:  biogeme.expressions.Expression.

        :return: numpy series, long as the number of entries
                 in the database, containing the calculated quantities.
        :rtype: numpy.Series

        :raise BiogemeError: if the database is empty.
        """

        if len(self.data.index) == 0:
            error_msg = 'Database has no entry'
            raise excep.BiogemeError(error_msg)

        return expression.getValue_c(database=self, prepareIds=True)

    def checkAvailabilityOfChosenAlt(self, avail, choice):
        """Check if the chosen alternative is available for each entry
        in the database.

        :param avail: list of expressions to evaluate the
                      availability conditions for each alternative.
        :type avail: list of biogeme.expressions.Expression
        :param choice: expression for the chosen alternative.
        :type choice: biogeme.expressions.Expression

        :return: numpy series of bool, long as the number of entries
                 in the database, containing True is the chosen alternative is
                 available, False otherwise.
        :rtype: numpy.Series

        :raise BiogemeError: if the chosen alternative does not appear
            in the availability dict
        :raise BiogemeError: if the database is empty.
        """
        self._avail = avail
        self._choice = choice

        if len(self.data.index) == 0:
            error_msg = 'Database has no entry'
            raise excep.BiogemeError(error_msg)

        choice_array = choice.getValue_c(
            database=self, aggregation=False, prepareIds=True
        )
        calculated_avail = {}
        for key, expression in avail.items():
            calculated_avail[key] = expression.getValue_c(
                database=self, aggregation=False, prepareIds=True
            )
        try:
            avail_chosen = np.array(
                [calculated_avail[c][i] for i, c in enumerate(choice_array)]
            )
        except KeyError as exc:
            for c in choice_array:
                err_msg = ''
                if c not in calculated_avail:
                    err_msg = (
                        f'Chosen alternative {c} does not appear in '
                        f'availability dict: {calculated_avail.keys()}'
                    )
                    raise excep.BiogemeError(err_msg) from exc
        return avail_chosen != 0

    def choiceAvailabilityStatistics(self, avail, choice):
        """Calculates the number of time an alternative is chosen and available

        :param avail: list of expressions to evaluate the
                      availability conditions for each alternative.
        :type avail: list of biogeme.expressions.Expression
        :param choice: expression for the chosen alternative.
        :type choice: biogeme.expressions.Expression

        :return: for each alternative, a tuple containing the number of time
            it is chosen, and the number of time it is available.
        :rtype: dict(int: (int, int))

        :raise BiogemeError: if the database is empty.
        """
        if len(self.data.index) == 0:
            error_msg = 'Database has no entry'
            raise excep.BiogemeError(error_msg)

        self._avail = avail
        self._choice = choice

        choice_array = choice.getValue_c(
            database=self,
            aggregation=False,
            prepareIds=True,
        )
        unique = np.unique(choice_array, return_counts=True)
        choice_stat = {alt: unique[1][i] for i, alt in enumerate(unique[0])}
        calculated_avail = {}
        for key, expression in avail.items():
            calculated_avail[key] = expression.getValue_c(
                database=self,
                aggregation=False,
                prepareIds=True,
            )
        avail_stat = {k: sum(a) for k, a in calculated_avail.items()}
        theResults = {alt: (c, avail_stat[alt]) for alt, c in choice_stat.items()}
        return theResults

    def scaleColumn(self, column, scale):
        """Multiply an entire column by a scale value

        :param column: name of the column
        :type column: string
        :param scale: value of the scale. All values of the column will
              be multiplied by that scale.
        :type scale: float

        """
        self.data[column] = self.data[column] * scale

    def suggestScaling(self, columns=None, reportAll=False):
        """Suggest a scaling of the variables in the database.

        For each column, :math:`\\delta` is the difference between the
        largest and the smallest value, or one if the difference is
        smaller than one. The level of magnitude is evaluated as a
        power of 10. The suggested scale is the inverse of this value.

        .. math:: s = \\frac{1}{10^{|\\log_{10} \\delta|}}

        where :math:`|x|` is the integer closest to :math:`x`.

        :param columns: list of columns to be considered.
                        If None, all of them will be considered.
        :type columns: list(str)

        :param reportAll: if False, remove entries where the suggested
            scale is 1, 0.1 or 10
        :type reportAll: bool

        :return: A Pandas dataframe where each row contains the name
                 of the variable and the suggested scale s. Ideally,
                 the column should be multiplied by s.

        :rtype: pandas.DataFrame

        :raise BiogemeError: if a variable in ``columns`` is unknown.
        """
        if columns is None:
            columns = self.data.columns
        else:
            for c in columns:
                if c not in self.data:
                    errorMsg = f'Variable {c} not found.'
                    raise excep.BiogemeError(errorMsg)

        largestValue = [
            max(np.abs(self.data[col].max()), np.abs(self.data[col].min()))
            for col in columns
        ]
        res = [
            [col, 1 / 10 ** np.round(np.log10(max(1.0, lv))), lv]
            for col, lv in zip(columns, largestValue)
        ]
        df = pd.DataFrame(res, columns=['Column', 'Scale', 'Largest'])
        if not reportAll:
            # Remove entries where the suggested scale is 1, 0.1 or 10
            remove = (df.Scale == 1) | (df.Scale == 0.1) | (df.Scale == 10)
            df.drop(df[remove].index, inplace=True)
        return df

    def sampleWithReplacement(self, size=None):
        """Extract a random sample from the database, with replacement.

        Useful for bootstrapping.

        :param size: size of the sample. If None, a sample of
               the same size as the database will be generated.
               Default: None.
        :type size: int

        :return: pandas dataframe with the sample.
        :rtype: pandas.DataFrame

        """
        if size is None:
            size = len(self.data)
        sample = self.data.iloc[np.random.randint(0, len(self.data), size=size)]
        return sample

    def sampleIndividualMapWithReplacement(self, size=None):
        """Extract a random sample of the individual map
        from a panel data database, with replacement.

        Useful for bootstrapping.

        :param size: size of the sample. If None, a sample of
                   the same size as the database will be generated.
                   Default: None.
        :type size: int

        :return: pandas dataframe with the sample.
        :rtype: pandas.DataFrame

        :raise BiogemeError: if the database in not in panel mode.
        """
        if not self.isPanel():
            errorMsg = (
                'Function sampleIndividualMapWithReplacement'
                ' is available only on panel data.'
            )
            raise excep.BiogemeError(errorMsg)

        if size is None:
            size = len(self.individualMap)
        sample = self.individualMap.iloc[
            np.random.randint(0, len(self.individualMap), size=size)
        ]
        return sample

    #####
    # This has to be reimplemented in a cleaner way
    ####
    #    def sampleWithoutReplacement(
    #        self, samplingRate, columnWithSamplingWeights=None
    #    ):
    #        """Replace the data set by a sample for stochastic algorithms
    #
    #        :param samplingRate: the proportion of data to include in the sample.
    #        :type samplingRate: float
    #        :param columnWithSamplingWeights: name of the column with
    #              the sampling weights. If None, each row has equal probability.
    #        :type columnWithSamplingWeights: string
    #
    #        :raise BiogemeError: if the structure of the database has been modified
    #            since last sample.
    #        """
    #        if self.isPanel():
    #            if self.fullIndividualMap is None:
    #                self.fullIndividualMap = self.individualMap
    #            # Check if the structure has not been modified since
    #            # last sample
    #            if set(self.fullIndividualMap.columns) != set(
    #                self.individualMap.columns
    #            ):
    #                message = (
    #                    'The structure of the database has been '
    #                    'modified since last sample. '
    #                )
    #                left = set(self.fullIndividualMap.columns).difference(
    #                    set(self.individualMap.columns)
    #                )
    #                if left:
    #                    message += f' Columns that disappeared: {left}'
    #                right = set(self.individualMap.columns).difference(
    #                    set(self.fullIndividualMap.columns)
    #                )
    #                if right:
    #                    message += f' Columns that were added: {right}'
    #                raise excep.BiogemeError(message)
    #
    #            self.individualMap = self.fullIndividualMap.sample(
    #                frac=samplingRate, weights=columnWithSamplingWeights
    #            )
    #        else:
    #            # Cross sectional data
    #            if self.fullData is None:
    #                self.fullData = self.data
    #            else:
    #                # Check if the structure has not been modified since
    #                # last sample
    #                if set(self.fullData.columns) != set(self.data.columns):
    #                    message = (
    #                        'The structure of the database has been modified '
    #                        'since last sample. '
    #                    )
    #                    left = set(self.fullData.columns).difference(
    #                        set(self.data.columns)
    #                    )
    #                    if left:
    #                        message += f' Columns that disappeared: {left}'
    #                    right = set(self.data.columns).difference(
    #                        set(self.fullData.columns)
    #                    )
    #                    if right:
    #                        message += f' Columns that were added: {right}'
    #                    raise excep.BiogemeError(message)
    #
    #            self.data = self.fullData.sample(
    #                frac=samplingRate, weights=columnWithSamplingWeights
    #            )

    #    def useFullSample(self):
    #        """Re-establish the full sample for calculation of the likelihood"""
    #        if self.isPanel():
    #            if self.fullIndividualMap is None:
    #                raise excep.BiogemeError(
    #                    'Full panel data set has not been saved.'
    #                )
    #            self.individualMap = self.fullIndividualMap
    #        else:
    #            if self.fullData is None:
    #                raise excep.BiogemeError('Full data set has not been saved.')
    #            self.data = self.fullData

    def addColumn(self, expression, column):
        """Add a new column in the database, calculated from an expression.

        :param expression:  expression to evaluate
        :type expression: biogeme.expressions.Expression

        :param column: name of the column to add
        :type column: string

        :return: the added column
        :rtype: numpy.Series

        :raises ValueError: if the column name already exists.
        :raise BiogemeError: if the database is empty.

        """
        if len(self.data.index) == 0:
            error_msg = 'Database has no entry'
            raise excep.BiogemeError(error_msg)

        if column in self.data.columns:
            raise ValueError(
                f'Column {column} already exists in the database {self.name}'
            )

        self._expression = expression
        new_column = self._expression.getValue_c(
            database=self, aggregation=False, prepareIds=True
        )
        self.data[column] = new_column
        self.variables[column] = Variable(column)
        return self.data[column]

    def DefineVariable(self, name, expression):
        """Insert a new column in the database and define it as a variable."""
        self.addColumn(expression, name)
        return Variable(name)

    def remove(self, expression):
        """Removes from the database all entries such that the value
        of the expression is not 0.

        :param expression: expression to evaluate
        :type expression: biogeme.expressions.Expression

        """
        columnName = '__bioRemove__'
        if isNumeric(expression):
            self.addColumn(Numeric(expression), columnName)
        else:
            self.addColumn(expression, columnName)
        self.excludedData = len(self.data[self.data[columnName] != 0].index)
        self.data.drop(self.data[self.data[columnName] != 0].index, inplace=True)
        self.data.drop(columns=[columnName], inplace=True)

    def check_segmentation(self, segmentation_tuple):
        """Check that the segmentation covers the complete database

        :param segmentation_tuple: object describing the segmentation
        :type segmentation_tuple: biogeme.segmentation.DiscreteSegmentationTuple

        :return: number of observations per segment.
        :rtype: dict(str: int)
        """

        all_values = self.data[segmentation_tuple.variable.name].value_counts()
        # Check if all values in the segmentation are in the database
        for value, name in segmentation_tuple.mapping.items():
            if value not in all_values:
                error_msg = (
                    f'Variable {segmentation_tuple.variable.name} does not '
                    f'take the value {value} representing segment "{name}"'
                )
                raise excep.BiogemeError(error_msg)
        for value, count in all_values.items():
            if value not in segmentation_tuple.mapping:
                error_msg = (
                    f'Variable {segmentation_tuple.variable.name} '
                    f'takes the value {value} [{count} times], and it does not '
                    f'define any segment.'
                )
                raise excep.BiogemeError(error_msg)

        named_values = {}
        for value, name in segmentation_tuple.mapping.items():
            named_values[name] = all_values[value]
        return named_values

    def dumpOnFile(self):
        """Dumps the database in a CSV formatted file.

        :return:  name of the file
        :rtype: string
        """
        theName = f'{self.name}_dumped'
        dataFileName = bf.getNewFileName(theName, 'dat')
        self.data.to_csv(dataFileName, sep='\t', index_label='__rowId')
        logger.info(f'File {dataFileName} has been created')
        return dataFileName

    def setRandomNumberGenerators(self, rng):
        """Defines user-defined random numbers generators.

        :param rng: a dictionary of generators. The keys of the dictionary
           characterize the name of the generators, and must be
           different from the pre-defined generators in Biogeme
           (see :func:`~biogeme.database.Database.generateDraws` for the list).
           The elements of the
           dictionary are functions that take two arguments: the
           number of series to generate (typically, the size of the
           database), and the number of draws per series.
        :type rng: dict

        Example::

            def logNormalDraws(sample_size, number_of_draws):
                return np.exp(np.random.randn(sample_size, number_of_draws))

            def exponentialDraws(sample_size, number_of_draws):
                return -1.0 * np.log(np.random.rand(sample_size, number_of_draws))

            # We associate these functions with a name
            dict = {'LOGNORMAL':(logNormalDraws,
                                 'Draws from lognormal distribution'),
                    'EXP':(exponentialDraws,
                           'Draws from exponential distributions')}
            myData.setRandomNumberGenerators(dict)

        :raise ValueError: if a reserved keyword is used for a
             user-defined draws.

        """
        for k in self.nativeRandomNumberGenerators:
            if k in rng:
                errorMsg = (
                    f'{k} is a reserved keyword for draws'
                    f' and cannot be used for user-defined '
                    f'generators'
                )
                raise ValueError(errorMsg)

        self.userRandomNumberGenerators = rng

    def generateDraws(self, types, names, number_of_draws):
        """Generate draws for each variable.


        :param types: A dict indexed by the names of the variables,
                      describing the types of draws. Each of them can
                      be a native type or any type defined by the
                      function
                      :func:`~biogeme.database.Database.setRandomNumberGenerators`.

                      Native types:

                      - ``'UNIFORM'``: Uniform U[0, 1],
                      - ``'UNIFORM_ANTI``: Antithetic uniform U[0, 1]',
                      - ``'UNIFORM_HALTON2'``: Halton draws with base 2,
                        skipping the first 10,
                      - ``'UNIFORM_HALTON3'``: Halton draws with base 3,
                        skipping the first 10,
                      - ``'UNIFORM_HALTON5'``: Halton draws with base 5,
                        skipping  the first 10,
                      - ``'UNIFORM_MLHS'``: Modified Latin Hypercube
                        Sampling on [0, 1],
                      - ``'UNIFORM_MLHS_ANTI'``: Antithetic Modified
                        Latin Hypercube Sampling on [0, 1],
                      - ``'UNIFORMSYM'``: Uniform U[-1, 1],
                      - ``'UNIFORMSYM_ANTI'``: Antithetic uniform U[-1, 1],
                      - ``'UNIFORMSYM_HALTON2'``: Halton draws on [-1, 1]
                        with base 2, skipping the first 10,
                      - ``'UNIFORMSYM_HALTON3'``: Halton draws on [-1, 1]
                        with base 3, skipping the first 10,
                      - ``'UNIFORMSYM_HALTON5'``: Halton draws on [-1, 1]
                        with base 5, skipping the first 10,
                      - ``'UNIFORMSYM_MLHS'``: Modified Latin Hypercube
                        Sampling on [-1, 1],
                      - ``'UNIFORMSYM_MLHS_ANTI'``: Antithetic Modified
                        Latin Hypercube Sampling on [-1, 1],
                      - ``'NORMAL'``: Normal N(0, 1) draws,
                      - ``'NORMAL_ANTI'``: Antithetic normal draws,
                      - ``'NORMAL_HALTON2'``: Normal draws from Halton
                        base 2 sequence,
                      - ``'NORMAL_HALTON3'``: Normal draws from Halton
                        base 3 sequence,
                      - ``'NORMAL_HALTON5'``: Normal draws from Halton
                        base 5 sequence,
                      - ``'NORMAL_MLHS'``: Normal draws from Modified
                        Latin Hypercube Sampling,
                      - ``'NORMAL_MLHS_ANTI'``: Antithetic normal draws
                        from Modified Latin Hypercube Sampling]

                      For an updated description of the native types, call the function
                      :func:`~biogeme.database.Database.descriptionOfNativeDraws`.



        :type types: dict

        :param names: the list of names of the variables that require draws
            to be generated.
        :type names: list of strings

        :param number_of_draws: number of draws to generate.
        :type number_of_draws: int

        :return: a 3-dimensional table with draws. The 3 dimensions are

              1. number of individuals
              2. number of draws
              3. number of variables

        :rtype: numpy.array

        Example::

              types = {'randomDraws1': 'NORMAL_MLHS_ANTI',
                       'randomDraws2': 'UNIFORM_MLHS_ANTI',
                       'randomDraws3': 'UNIFORMSYM_MLHS_ANTI'}
              theDrawsTable = myData.generateDraws(types,
                  ['randomDraws1', 'randomDraws2', 'randomDraws3'], 10)


        :raise BiogemeError: if a type of draw is unknown.

        :raise BiogemeError: if the output of the draw generator does not
            have the requested dimensions.

        """
        self.number_of_draws = number_of_draws
        # Dimensions of the draw table:
        # 1. number of variables
        # 2. number of individuals
        # 3. number of draws
        listOfDraws = [None] * len(names)
        for i, v in enumerate(names):
            name = v
            drawType = types[name]
            self.typesOfDraws[name] = drawType
            theGenerator = self.nativeRandomNumberGenerators.get(drawType)
            if theGenerator is None:
                theGenerator = self.userRandomNumberGenerators.get(drawType)
                if theGenerator is None:
                    native = self.nativeRandomNumberGenerators
                    user = self.userRandomNumberGenerators
                    errorMsg = (
                        f'Unknown type of draws for '
                        f'variable {name}: {drawType}. '
                        f'Native types: {native}. '
                        f'User defined: {user}'
                    )
                    raise excep.BiogemeError(errorMsg)
            listOfDraws[i] = theGenerator[0](self.getSampleSize(), number_of_draws)
            if listOfDraws[i].shape != (self.getSampleSize(), number_of_draws):
                errorMsg = (
                    f'The draw generator for {name} must'
                    f' generate a numpy array of dimensions'
                    f' ({self.getSampleSize()}, {number_of_draws})'
                    f' instead of {listOfDraws[i].shape}'
                )
                raise excep.BiogemeError(errorMsg)

        self.theDraws = np.array(listOfDraws)
        # Draws as a three-dimensional numpy series. The dimensions
        # are organized to be more suited for calculation.
        # 1. number of individuals
        # 2. number of draws
        # 3. number of variables
        self.theDraws = np.moveaxis(self.theDraws, 0, -1)
        return self.theDraws

    def getNumberOfObservations(self):
        """
        Reports the number of observations in the database.

        Note that it returns the same value, irrespectively
        if the database contains panel data or not.

        :return: Number of observations.
        :rtype: int

        See also:  getSampleSize()
        """
        return self.data.shape[0]

    def getSampleSize(self):
        """Reports the size of the sample.

        If the data is cross-sectional, it is the number of
        observations in the database. If the data is panel, it is the
        number of individuals.

        :return: Sample size.
        :rtype: int

        See also: getNumberOfObservations()

        """
        if self.isPanel():
            return self.individualMap.shape[0]

        return self.data.shape[0]

    def split(self, slices, groups=None):
        """Prepare estimation and validation sets for validation.

        :param slices: number of slices
        :type slices: int

        :param groups: name of the column that defines the ID of the
            groups. Data belonging to the same groups will be maintained
            together.
        :type groups: str

        :return: list of estimation and validation data sets
        :rtype: list(tuple(pandas.DataFrame, pandas.DataFrame))

        :raise BiogemeError: if the number of slices is less than two

        """
        if slices < 2:
            error_msg = (
                f'The number of slices is {slices}. It must be greater '
                f'or equal to 2.'
            )
            raise excep.BiogemeError(error_msg)

        if groups is not None and self.isPanel():
            if groups != self.panelColumn:
                error_msg = (
                    f'The data is already organized by groups on '
                    f'{self.panelColumn}. The grouping by {groups} '
                    f'cannot be done.'
                )
                raise excep.BiogemeError(error_msg)

        if self.isPanel():
            groups = self.panelColumn

        if groups is None:
            shuffled = self.data.sample(frac=1)
            theSlices = np.array_split(shuffled, slices)
        else:
            ids = self.data[groups].unique()
            np.random.shuffle(ids)
            the_slices_ids = np.array_split(ids, slices)
            theSlices = [
                self.data[self.data[groups].isin(ids)] for ids in the_slices_ids
            ]
        estimationSets = []
        validationSets = []
        for i, v in enumerate(theSlices):
            estimationSets.append(pd.concat(theSlices[:i] + theSlices[i + 1 :]))
            validationSets.append(v)
        for e, v in zip(estimationSets, validationSets):
            print(type(e))
            print(type(v))
        return [
            EstimationValidation(estimation=e, validation=v)
            for e, v in zip(estimationSets, validationSets)
        ]

    def isPanel(self):
        """Tells if the data is panel or not.

        :return: True if the data is panel.
        :rtype: bool
        """
        return self.panelColumn is not None

    def panel(self, columnName):
        """Defines the data as panel data

        :param columnName: name of the columns that identifies individuals.
        :type columnName: string

        :raise BiogemeError: if the data are not sorted properly, that
            is if the data for the one individuals are not consecutive.

        """

        self.panelColumn = columnName

        # Check if the data is organized in consecutive entries
        # Number of groups of data
        nGroups = tools.countNumberOfGroups(self.data, self.panelColumn)
        sortedData = self.data.sort_values(by=[self.panelColumn])
        nIndividuals = tools.countNumberOfGroups(sortedData, self.panelColumn)
        if nGroups != nIndividuals:
            theError = (
                f'The data must be sorted so that the data'
                f' for the same individual are consecutive.'
                f' There are {nIndividuals} individuals '
                f'in the sample, and {nGroups} groups of '
                f'data for column {self.panelColumn}.'
            )
            raise excep.BiogemeError(theError)

        self.buildPanelMap()

    def buildPanelMap(self):
        """Sorts the data so that the observations for each individuals are
        contiguous, and builds a map that identifies the range of indices of
        the observations of each individuals.
        """
        if self.panelColumn is not None:
            self.data = self.data.sort_values(by=self.panelColumn)
            # It is necessary to renumber the row to reflect the new ordering
            self.data.index = range(len(self.data.index))
            local_map = {}
            individuals = self.data[self.panelColumn].unique()
            for i in individuals:
                indices = self.data.loc[self.data[self.panelColumn] == i].index
                local_map[i] = [min(indices), max(indices)]
            self.individualMap = pd.DataFrame(local_map).T
            self.fullIndividualMap = self.individualMap

    def count(self, columnName, value):
        """Counts the number of observations that have a specific value in a
        given column.

        :param columnName: name of the column.
        :type columnName: string
        :param value: value that is seeked.
        :type value: float

        :return: Number of times that the value appears in the column.
        :rtype: int
        """
        return self.data[self.data[columnName] == value].count()[columnName]

    def generateFlatPanelDataframe(self, saveOnFile=None, identical_columns=tuple()):
        """Generate a flat version of the panel data

        :param saveOnFile: if True, the flat database is saved on file.
        :type saveOnFile: bool

        :param identical_columns: tuple of columns that contain the
            same values for all observations of the same
            individual. Default: empty list.

        :type identical_columns: tuple(str)

        :return: the flatten database, in Pandas format
        :rtype: pandas.DataFrame

        :raise BiogemeError: if the database in not panel

        """
        if not self.isPanel():
            error_msg = 'This function can only be called for panel data'
            raise excep.BiogemeError(error_msg)
        flat_data = tools.flatten_database(
            self.data, self.panelColumn, identical_columns=identical_columns
        )
        if saveOnFile:
            file_name = f'{self.name}_flatten.csv'
            flat_data.to_csv(file_name)
            logger.info(f'File {file_name} has been created.')
        return flat_data

    def __str__(self):
        """Allows to print the dabase"""
        result = f'biogeme database {self.name}:\n{self.data}'
        if self.isPanel():
            result += f'\nPanel data\n{self.individualMap}'
        return result

    def verify_segmentation(self, segmentation):
        """Verifies if the definition of the segmentation is consistent with the data

        :param segmentation: definition of the segmentation
        :type segmentation: DiscreteSegmentationTuple

        :raise BiogemeError: if the segmentation is not consistent with the data.
        """

        variable = (
            segmentation.variable
            if isinstance(segmentation.variable, Variable)
            else Variable(segmentation.variable)
        )

        # Check if the variable is in the database.
        if variable.name not in self.data.columns:
            error_msg = f'Unknown variable {variable.name}'
            raise excep.BiogemeError(error_msg)

        # Extract all unique values from the data base.
        unique_values = set(self.data[variable.name].unique())
        segmentation_values = set(segmentation.mapping.keys())

        in_data_not_in_segmentation = unique_values - segmentation_values
        in_segmentation_not_in_data = segmentation_values - unique_values

        error_msg_1 = (
            (
                f'The following entries are missing in the segmentation: '
                f'{in_data_not_in_segmentation}.'
            )
            if in_data_not_in_segmentation
            else ''
        )

        error_msg_2 = (
            (
                f'Segmentation entries do not exist in the data: '
                f'{in_segmentation_not_in_data}.'
            )
            if in_segmentation_not_in_data
            else ''
        )

        if error_msg_1 or error_msg_2:
            raise excep.BiogemeError(f'{error_msg_1} {error_msg_2}')

    def generate_segmentation(self, variable, mapping=None, reference=None):
        """Generate a segmentation tuple for a variable.

        :param variable: Variable object or name of the variable
        :type variable: biogeme.expressions.Variable or string

        :param mapping: mapping associating values of the variable to
            names. If incomplete, default names are provided.
        :type mapping: dict(int: str)

        :param reference: name of the reference category. If None, an
            arbitrary category is selected as reference.  :type:
        :type reference: str


        """

        the_variable = (
            variable if isinstance(variable, Variable) else Variable(variable)
        )

        # Check if the variable is in the database.
        if the_variable.name not in self.data.columns:
            error_msg = f'Unknown the_variable {the_variable.name}'
            raise excep.BiogemeError(error_msg)

        # Extract all unique values from the data base.
        unique_values = set(self.data[the_variable.name].unique())

        if len(unique_values) >= 10:
            warning_msg = (
                f'Variable {the_variable.name} takes a total of '
                f'{len(unique_values)} different values in the database. It is '
                f'likely to be too large for a discrete segmentation.'
            )
            logger.warning(warning_msg)

        # Check that the provided mapping is consistent with the data
        values_not_in_data = [
            value for value in mapping.keys() if value not in unique_values
        ]

        if values_not_in_data:
            error_msg = (
                f'The following values in the mapping do not exist in the data for '
                f'variable {the_variable.name}: {values_not_in_data}'
            )
            raise excep.BiogemeError(error_msg)

        the_mapping = {value: f'{the_variable.name}_{value}' for value in unique_values}

        if mapping is not None:
            the_mapping.update(mapping)

        if reference is not None and reference not in mapping.values():
            error_msg = (
                f'Level {reference} of variable {the_variable.name} does not '
                'appear in the mapping: {mapping.values()}'
            )
            raise excep.BiogemeError(error_msg)

        return DiscreteSegmentationTuple(
            variable=the_variable,
            mapping=the_mapping,
            reference=reference,
        )



"""Centralized management of the parameters.

Parameters are organized by section. Parameters within the same
section must have different names

:author: Michel Bierlaire
:date: Fri Dec  2 12:17:38 2022

"""

import logging
from datetime import datetime
from typing import NamedTuple
import tomlkit as tk
import biogeme.exceptions as excep
from biogeme.version import getVersion
from biogeme.default_parameters import all_parameters_tuple, ParameterTuple

logger = logging.getLogger(__name__)

TRUE_STR = ('True', 'true', 'Yes', 'yes')
FALSE_STR = ('False', 'false', 'No', 'no')
DEFAULT_FILE_NAME = 'biogeme.toml'
ROW_LENGTH = 80


class NameSectionTuple(NamedTuple):
    name: str
    section: str


def format_comment(the_param):
    """Format the content of the file, in particular the description of the parameter

    :param the_param: parmeter to format
    :type the_param: biogeme.default_parameters.ParameterTuple

    :return: formatted text
    :rtype: str
    """
    the_str = f'{the_param.name} = {the_param.value}'
    multiplier = len(the_str) + 1
    words = the_param.description.split(' ')
    current_row = '#'
    rows = []
    for word in words:
        the_length = len(current_row)
        if not rows:
            the_length += multiplier
        if the_length + len(word) > ROW_LENGTH:
            # Start a new line
            rows.append(current_row)
            blanks = ' ' * multiplier
            if the_param.type in (bool, str):
                blanks += '  '
            current_row = f'{blanks}#'
        current_row += ' ' + word
    if current_row != '':
        rows.append(current_row)
    formatted = '\n'.join(rows)
    return formatted


def parse_boolean(value):
    """Transforms one of the string representing a boolean into an actual boolean

    :param value: value to be transformed
    :type value: str

    :return: equivalent boolean
    :rtype: bool

    """
    if value in TRUE_STR:
        return True
    if value in FALSE_STR:
        return False

    error_msg = f'{value} is not a valid boolean. Use "True" of "False"'
    raise excep.BiogemeError(error_msg)


class Parameters:
    """Parameters management"""

    def __init__(self):

        # Store values in a dict
        self.document = None  # TOML document
        self.file_name = None
        self.all_parameters_dict = {}
        # Add the default parameters
        for param in all_parameters_tuple:
            self.add_parameter(param)

    def add_parameter(self, parameter_tuple):
        """Add one aprameter

        :param parameter_tuple: tuple containing the data associated with the parameter
        :type parameter_tuple: biogeme.default_parameters.ParameterTuple
        """
        key = NameSectionTuple(
            name=parameter_tuple.name, section=parameter_tuple.section
        )
        ok, messages = self.check_parameter_value(parameter_tuple)
        if not ok:
            raise excep.BiogemeError(messages)
        self.all_parameters_dict[key] = parameter_tuple

    def read_file(self, file_name):
        """Read TOML file"""
        self.file_name = DEFAULT_FILE_NAME if file_name is None else file_name
        try:
            with open(self.file_name, 'r', encoding='utf-8') as f:
                content = f.read()
                self.document = tk.parse(content)
                self.import_document()
                logger.info(f'File {self.file_name} has been parsed.')

        except FileNotFoundError:
            self.dump_file(self.file_name)

    def parameters_in_section(self, section_name):
        """Returns the names of the parameters in a section

        :param section_name: name of the section
        :type section_name: str

        """
        results = [
            p.name
            for p in self.all_parameters_dict.values()
            if p.section == section_name
        ]
        return results

    def sections_from_name(self, name):
        """Obtain the list of sections containing the given entry

        :param name: name of the parameter
        :type name: str
        """
        return {p.section for p in self.all_parameters_dict if p.name == name}

    def dump_file(self, file_name):
        """Dump the values of the parameters in the TOML file"""
        self.document = self.generate_document()
        with open(file_name, 'w', encoding='utf-8') as f:
            print(tk.dumps(self.document), file=f)
        logger.warning(f'File {file_name} has been created')

    def import_document(self):
        """Record the values of the parameters in the TOML document"""
        for section_name, entries in self.document.items():
            for entry_name, entry_value in entries.items():
                key = NameSectionTuple(name=entry_name, section=section_name)
                default = self.all_parameters_dict.get(key)
                if default is None:
                    warning_msg = (
                        f'Entry {entry_name} in Section {section_name} is '
                        f'ignored by Biogeme.'
                    )
                    logger.warning(warning_msg)
                    continue
                if entry_value is None:
                    value = default.value
                elif default.type is bool:
                    try:
                        value = parse_boolean(entry_value)
                    except excep.BiogemeError as e:
                        error_msg = (
                            f'Error with entry {entry_name} in Section '
                            f'{section_name}: {e}'
                        )
                        raise error_msg from e
                else:
                    value = entry_value

                the_parameter = ParameterTuple(
                    name=entry_name,
                    value=value,
                    type=default.type,
                    section=section_name,
                    description=default.description,
                    check=default.check,
                )
                self.add_parameter(the_parameter)

    def generate_document(self):
        """Generate the  TOML document"""
        current_datetime = datetime.now()
        formatted_datetime = current_datetime.strftime("%B %d, %Y. %H:%M:%S")
        doc = tk.document()

        doc.add(tk.comment(f'Default parameter file for Biogeme {getVersion()}'))
        doc.add(tk.comment(f'Automatically created on {formatted_datetime}'))

        sections = {p.section for p in self.all_parameters_dict}

        tables = {section: tk.table() for section in sections}

        for parameter in self.all_parameters_dict.values():
            if isinstance(parameter.value, bool):
                value = 'True' if parameter.value else 'False'
            else:
                value = parameter.value
            tables[parameter.section].add(parameter.name, value)
            tables[parameter.section][parameter.name].comment(format_comment(parameter))

        for s, t in tables.items():
            doc[s] = t
        return doc

    def get_param_tuple(self, name, section=None):
        """Obtain a tuple describing the parameter

        :param name: name of the parameter
        :type name: str

        :param section: section where the parameter is relevant
        :type section: str

        :return: tuple containing the name, section and value of the
            parameter, or None if not found
        :rtype: biogeme.default_parameters.ParameterTuple

        """
        if section is None:
            sections = self.sections_from_name(name)
            if len(sections) > 1:
                error_msg = (
                    f'Ambiguity: parameter {name} belongs to multiple '
                    f'sections: {sections}.'
                )
                raise excep.BiogemeError(error_msg)
            if not sections:
                error_msg = f'Parameter {name} does not belong to any section.'
                raise excep.BiogemeError(error_msg)
            section = sections.pop()

        name_section = NameSectionTuple(name=name, section=section)
        the_tuple = self.all_parameters_dict.get(name_section)
        if the_tuple is None:
            if section is None:
                error_msg = f'Unknown parameter {name}'
            else:
                known = self.parameters_in_section(section)
                if known:
                    known_msg = f'Known parameter(s): {", ".join(known)}.'
                else:
                    known_msg = 'The section does not exist.'
                error_msg = (
                    f'Unknown parameter {name} in section {section}. {known_msg}'
                )
            raise excep.BiogemeError(error_msg)

        return the_tuple

    @staticmethod
    def check_parameter_value(the_tuple):
        """Check if the value of the parameter is valid

        :param the_tuple: tuple to check
        :type the_tuple: biogeme.default_parameters.ParameterTuple

        :return: a boolean that is True if the value is valid. If not,
            diagnostics are provided.
        :rtype: tuple(bool, list(str))
        """
        ok = True
        messages = []
        if the_tuple.check is None:
            return ok, messages
        for check in the_tuple.check:
            is_ok, diag = check(the_tuple.value)
            if not is_ok:
                ok = False
                messages.append(
                    f'Error for parameter {the_tuple.name} in '
                    f'section {the_tuple.section}. {diag}'
                )
        return ok, messages

    def set_value(self, name, value, section=None):
        """Set a value of a parameter. If the parameter appears in
        only one section, the name of the section can be ommitted.

        :param name: name of the parameter
        :type name: str

        :param value: value of the parameter
        :type value: str

        :param section: section containing the parameter
        :type section: str
        """
        the_tuple = self.get_param_tuple(name, section)
        the_parameter = ParameterTuple(
            name=the_tuple.name,
            value=value,
            type=the_tuple.type,
            section=the_tuple.section,
            description=the_tuple.description,
            check=the_tuple.check,
        )

        self.add_parameter(the_parameter)

    def get_value(self, name, section=None):
        """Get the value of a parameter. If the parameter appears in
        only one section, the name of the section can be ommitted.

        :param name: name of the parameter
        :type name: str

        :param section: section containing the parameter
        :type section: str
        """
        the_tuple = self.get_param_tuple(name, section)
        # logger.info(f'{name} = {the_tuple.value}')
        return the_tuple.value


biogeme_parameters = Parameters()



"""
Test the optimization module

:author: Michel Bierlaire
:data: Wed Apr 29 17:45:19 2020
"""
# Bug in pylint
# pylint: disable=no-member
#
# Too constraining
# pylint: disable=invalid-name
#
# Not needed in test
# pylint: disable=missing-function-docstring, missing-class-docstring


import os
import shutil
import unittest
import random as rnd

import numpy as np
import tempfile

import biogeme.biogeme as bio
from biogeme import models
import biogeme.optimization as opt
import biogeme.exceptions as excep
from biogeme.expressions import Variable, Beta
from test_data import getData


class test_optimization(unittest.TestCase):
    def setUp(self):
        np.random.seed(90267)
        rnd.seed(90267)
        Choice = Variable('Choice')
        Variable1 = Variable('Variable1')
        Variable2 = Variable('Variable2')
        beta1 = Beta('beta1', 0, None, None, 0)
        beta2 = Beta('beta2', 0, None, None, 0)
        V1 = beta1 * Variable1
        V2 = beta2 * Variable2
        V3 = 0
        V = {1: V1, 2: V2, 3: V3}

        self.likelihood = models.loglogit(V, av=None, i=Choice)

        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        self.scipy_file = os.path.join(self.test_dir, 'scipy.toml')
        with open(self.scipy_file, 'w', encoding='utf-8') as f:
            print('[Estimation]', file=f)
            print('optimization_algorithm = "scipy"', file=f)
        self.ls_file = os.path.join(self.test_dir, 'line_search.toml')
        with open(self.ls_file, 'w', encoding='utf-8') as f:
            print('[Estimation]', file=f)
            print('optimization_algorithm = "LS-newton"', file=f)
        self.tr_file = os.path.join(self.test_dir, 'trust_region.toml')
        with open(self.tr_file, 'w', encoding='utf-8') as f:
            print('[Estimation]', file=f)
            print('optimization_algorithm = "TR-newton"', file=f)
        self.simple_bounds_file = os.path.join(self.test_dir, 'simple_bounds.toml')
        with open(self.simple_bounds_file, 'w', encoding='utf-8') as f:
            print('[Estimation]', file=f)
            print('optimization_algorithm = "simple_bounds"', file=f)

    def tearDown(self):
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)

    def testBioScipy(self):
        my_biogeme = bio.BIOGEME(
            getData(1), self.likelihood, parameter_file=self.scipy_file
        )
        my_biogeme.modelName = 'simpleExample'
        my_biogeme.generateHtml = False
        my_biogeme.generatePickle = False
        my_biogeme.saveIterations = False
        results = my_biogeme.estimate()
        beta = results.getBetaValues()
        self.assertAlmostEqual(beta['beta1'], 0.144546, 3)
        self.assertAlmostEqual(beta['beta2'], 0.023502, 3)

    def testBioNewtonLineSearch(self):
        my_biogeme = bio.BIOGEME(
            getData(1), self.likelihood, parameter_file=self.ls_file
        )
        my_biogeme.modelName = 'simpleExample'
        my_biogeme.generateHtml = False
        my_biogeme.generatePickle = False
        my_biogeme.saveIterations = False
        results = my_biogeme.estimate()
        beta = results.getBetaValues()
        self.assertAlmostEqual(beta['beta1'], 0.144546, 3)
        self.assertAlmostEqual(beta['beta2'], 0.023502, 3)

    def testBioNewtonTrustRegion(self):
        my_biogeme = bio.BIOGEME(
            getData(1), self.likelihood, parameter_file=self.tr_file
        )
        my_biogeme.modelName = 'simpleExample'
        my_biogeme.generateHtml = False
        my_biogeme.generatePickle = False
        my_biogeme.saveIterations = False
        results = my_biogeme.estimate()
        beta = results.getBetaValues()
        self.assertAlmostEqual(beta['beta1'], 0.144546, 3)
        self.assertAlmostEqual(beta['beta2'], 0.023502, 3)

    def testBioNewtonSimpleBounds(self):
        my_biogeme = bio.BIOGEME(
            getData(1), self.likelihood, parameter_file=self.simple_bounds_file
        )
        my_biogeme.modelName = 'simpleExample'
        my_biogeme.generateHtml = False
        my_biogeme.generatePickle = False
        my_biogeme.saveIterations = False
        results = my_biogeme.estimate()
        beta = results.getBetaValues()
        self.assertAlmostEqual(beta['beta1'], 0.144546, 3)
        self.assertAlmostEqual(beta['beta2'], 0.023502, 3)


if __name__ == '__main__':
    unittest.main()



"""File b22process_pareto.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 17:25:41 2023

The assisted specification algorithm generates a file containg the
pareto optimal specification. This script is designed to re-estimate
the Pareto optimal models

"""
try: 
    import matplotlib.pyplot as plt
    can_plot = True
except ModuleNotFoundError:
    can_plot = False
from biogeme.assisted import ParetoPostProcessing
from biogeme.results import compileEstimationResults
from b22multiple_models_spec import the_biogeme, PARETO_FILE_NAME

CSV_FILE = 'b22process_pareto.csv'
SEP_CSV = ','

the_pareto_post = ParetoPostProcessing(
    biogeme_object=the_biogeme,
    pareto_file_name=PARETO_FILE_NAME,
)
the_pareto_post.log_statistics()

all_results = the_pareto_post.reestimate(recycle=True)

summary, description = compileEstimationResults(all_results, use_short_names=True)

print(summary)
print(f'Summary table available in {CSV_FILE}')
summary.to_csv(CSV_FILE, sep=SEP_CSV)
with open(CSV_FILE, 'a', encoding='utf-8') as f:
    print('\n\n', file=f)
    for k, v in description.items():
        if k != v:
            print(f'{k}: {v}')
            print(f'{k}{SEP_CSV}{v}', file=f)


if can_plot:
    _ = the_pareto_post.plot()
    plt.show()



"""File b11cnl_wtp.py

:author: Michel Bierlaire, EPFL
:date: Fri Apr 14 16:48:29 2023

 Example of a cross-nested logit model with moneymetric utilities (WTP space)

"""

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta

from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

value_of_car = Beta('value_of_car', 0, None, None, 0)
value_of_train = Beta('value_of_train', 0, None, None, 0)
value_of_time = Beta('value_of_time', 0, None, None, 0)

mu = Beta('mu', 1, 0, None, 0)

mu_existing = Beta('mu_existing', 1, 1, None, 0)
mu_public = Beta('mu_public', 1, 1, None, 0)
alpha_existing = Beta('alpha_existing', 0.5, 0, 1, 0)
alpha_public = 1 - alpha_existing

# Definition of the cost functions
cost_1 = TRAIN_COST_SCALED + value_of_train + value_of_time * TRAIN_TT_SCALED
cost_2 = SM_COST_SCALED + value_of_time * SM_TT_SCALED
cost_3 = CAR_CO_SCALED + value_of_car + value_of_time * CAR_TT_SCALED

# To obtain the utilities, we change the sign of the cost
# functions. Note that we do not scale the utilities here, as the
# scale parameter is involved in the MEV function of the nested logit
# model

# To obtain the utilities, we change the sign of the cost
# functions. We also scale them with the parameter mu

scaled_V1 = -mu * cost_1
scaled_V2 = -mu * cost_2
scaled_V3 = -mu * cost_3

# Associate utility functions with the numbering of alternatives
V = {1: scaled_V1, 2: scaled_V2, 3: scaled_V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of nests
# Nest membership parameters
alpha_existing_assign = {1: alpha_existing, 2: 0.0, 3: 1.0}

alpha_public_assign = {1: alpha_public, 2: 1.0, 3: 0.0}

nest_existing = mu_existing, alpha_existing_assign
nest_public = mu_public, alpha_public_assign
nests = nest_existing, nest_public

# The choice model is a cross-nested logit, with availability conditions
logprob = models.logcnl_avail(V, av, nests, CHOICE)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b11cnl_wpt'

# Estimate the parameters
results = the_biogeme.estimate()
print(results.shortSummary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)
